from __future__ import annotations

import numpy as np

from shape_kernel import Cp, LocalFrame
from shape_kernel._types import as_np3, as_vec3, normalized

from .icosahedral_face import IcosehedralFace
from .quasi_tile import QuasiTile


class QuasiTile_02(QuasiTile):
    def __init__(self, oFrame0c: LocalFrame, fFaceSide: float = 20.0) -> None:
        super().__init__()
        self.m_clr = Cp.clrLemongrass

        s_ref_face = IcosehedralFace(oFrame0c, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.LINE, fFaceSide)
        f_long_axis = float(np.linalg.norm(as_np3(s_ref_face.vecPt3) - as_np3(s_ref_face.vecPt1)))

        o_frame_1b = oFrame0c
        o_frame_1t = LocalFrame.oGetTranslatedFrame(oFrame0c, tuple(f_long_axis * x for x in oFrame0c.vecGetLocalZ()))
        o_frame_1t = LocalFrame.oGetInvertFrame(o_frame_1t, True, True)

        s_face_1b = IcosehedralFace(o_frame_1b, IcosehedralFace.EDef.CENTRE, IcosehedralFace.EConnector.LINE, fFaceSide)
        s_face_1t = IcosehedralFace(o_frame_1t, IcosehedralFace.EDef.CENTRE, IcosehedralFace.EConnector.LINE, fFaceSide)

        a_side_faces: list[IcosehedralFace] = []
        a_angled_faces: list[IcosehedralFace] = []

        for vec_side_1b, vec_side_1t in ((s_face_1b.vecPt2, s_face_1t.vecPt2), (s_face_1b.vecPt4, s_face_1t.vecPt4)):
            vec_long_1s = as_vec3(normalized(as_np3(vec_side_1t) - as_np3(vec_side_1b)))
            vec_short_1s = as_vec3(normalized(as_np3(s_face_1b.vecPt3) - as_np3(s_face_1b.vecPt1)))
            vec_normal_1s = as_vec3(np.cross(as_np3(vec_long_1s), as_np3(vec_short_1s)))
            o_frame_1s = LocalFrame(vec_side_1b, vec_normal_1s, vec_long_1s)
            s_face_1s = IcosehedralFace(o_frame_1s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.TRIANGLE, fFaceSide)
            a_side_faces.append(s_face_1s)

            vec_centre_2s = as_vec3(0.5 * (as_np3(s_face_1b.vecPt1) + as_np3(s_face_1s.vecPt2)))
            vec_long_2s = as_vec3(normalized(as_np3(vec_centre_2s) - as_np3(vec_side_1b)))
            vec_short_2s = as_vec3(normalized(as_np3(vec_centre_2s) - as_np3(s_face_1b.vecPt1)))
            vec_normal_2s = as_vec3(np.cross(as_np3(vec_long_2s), as_np3(vec_short_2s)))
            o_frame_2s = LocalFrame(vec_side_1b, vec_normal_2s, vec_long_2s)
            s_face_2s = IcosehedralFace(o_frame_2s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.ARROW, fFaceSide)

            vec_centre_3s = as_vec3(0.5 * (as_np3(s_face_1b.vecPt3) + as_np3(s_face_1s.vecPt4)))
            vec_long_3s = as_vec3(normalized(as_np3(vec_centre_3s) - as_np3(vec_side_1b)))
            vec_short_3s = as_vec3(normalized(as_np3(vec_centre_3s) - as_np3(s_face_1b.vecPt3)))
            vec_normal_3s = as_vec3(np.cross(as_np3(vec_long_3s), as_np3(vec_short_3s)))
            o_frame_3s = LocalFrame(vec_side_1b, vec_normal_3s, vec_long_3s)
            s_face_3s = IcosehedralFace(o_frame_3s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.ARROW, fFaceSide)

            a_angled_faces.extend([s_face_2s, s_face_3s])

            vec_centre_2s = as_vec3(0.5 * (as_np3(s_face_1t.vecPt1) + as_np3(s_face_1s.vecPt4)))
            vec_long_2s = as_vec3(normalized(as_np3(vec_centre_2s) - as_np3(vec_side_1t)))
            vec_short_2s = as_vec3(normalized(as_np3(vec_centre_2s) - as_np3(s_face_1t.vecPt1)))
            vec_normal_2s = as_vec3(np.cross(as_np3(vec_long_2s), as_np3(vec_short_2s)))
            o_frame_2s = LocalFrame(vec_side_1t, vec_normal_2s, vec_long_2s)
            s_face_2s = IcosehedralFace(o_frame_2s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.TRIANGLE, fFaceSide)

            vec_centre_3s = as_vec3(0.5 * (as_np3(s_face_1t.vecPt3) + as_np3(s_face_1s.vecPt2)))
            vec_long_3s = as_vec3(normalized(as_np3(vec_centre_3s) - as_np3(vec_side_1t)))
            vec_short_3s = as_vec3(normalized(as_np3(vec_centre_3s) - as_np3(s_face_1t.vecPt3)))
            vec_normal_3s = as_vec3(np.cross(as_np3(vec_long_3s), as_np3(vec_short_3s)))
            o_frame_3s = LocalFrame(vec_side_1t, vec_normal_3s, vec_long_3s)
            s_face_3s = IcosehedralFace(o_frame_3s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.TRIANGLE, fFaceSide)

            a_angled_faces.extend([s_face_2s, s_face_3s])

        for i in range(len(a_side_faces)):
            a_side_faces[i].FlipAroundShortAxis()
            a_side_faces[i].FlipAroundLongAxis()

        self.m_aFaces = [s_face_1b, s_face_1t]
        self.m_aFaces.extend(a_side_faces)
        self.m_aFaces.extend(a_angled_faces)


__all__ = ["QuasiTile_02"]



from .cylindrical_control_spline import CylindricalControlSpline
from .control_point_spline import ControlPointSpline
from .tangent_control_spline import TangentialControlSpline
from .control_surface import ControlPointSurface
from .i_spline import ISpline

__all__ = [
    "ISpline",
    "ControlPointSpline",
    "ControlPointSurface",
    "CylindricalControlSpline",
    "TangentialControlSpline",
]



from .types import ColorLike, Vector2Like, Vector3Like, as_np3, as_tuple3, clamp01, normalized, np_vec3

__all__ = [
    "ColorLike",
    "Vector2Like",
    "Vector3Like",
    "as_np3",
    "as_tuple3",
    "clamp01",
    "normalized",
    "np_vec3",
]


from __future__ import annotations

import math

from picogk import Mesh, Voxels

from .._types import Vec3, as_np3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from ..modulations import GenericContour, LineModulation
from ..utils import VecOperations
from .base_shape import BaseShape


class BaseRevolve(BaseShape):
	def __init__(self, oFrame: LocalFrame, aFrames: Frames, fInwardRadius: float = 3.0, fOutwardRadius: float = 3.0) -> None:
		super().__init__()
		self.m_oFrame = oFrame
		self.m_aFrames = aFrames
		self.SetRadialSteps(100)
		self.SetPolarSteps(360)
		self.SetLengthSteps(500)
		self.m_oOuterRadiusModulation = LineModulation(fOutwardRadius)
		self.m_oInnerRadiusModulation = LineModulation(fInwardRadius)

	def SetRadius(self, oInnerRadiusOverCylinder: LineModulation, oOuterRadiusOverCylinder: LineModulation) -> None:
		self.m_oInnerRadiusModulation = oInnerRadiusOverCylinder
		self.m_oOuterRadiusModulation = oOuterRadiusOverCylinder

	def SetRadialSteps(self, nRadialSteps: int) -> None:
		self.m_nRadialSteps = max(5, int(nRadialSteps))

	def SetPolarSteps(self, nPolarSteps: int) -> None:
		self.m_nPolarSteps = max(5, int(nPolarSteps))

	def SetLengthSteps(self, nLengthSteps: int) -> None:
		self.m_nLengthSteps = max(5, int(nLengthSteps))

	def fGetInnerRadius(self, fLengthRatio: float) -> float:
		return self.m_oInnerRadiusModulation.fGetModulation(fLengthRatio)

	def fGetOuterRadius(self, fLengthRatio: float) -> float:
		return self.m_oOuterRadiusModulation.fGetModulation(fLengthRatio)

	def voxConstruct(self) -> Voxels:
		return Voxels.from_mesh(self.mshConstruct())

	def mshConstruct(self) -> Mesh:
		mesh = Mesh()
		for ip in range(1, self.m_nPolarSteps):
			p1, p2 = self._phi_ratio(ip - 1), self._phi_ratio(ip)
			for ir in range(1, self.m_nRadialSteps):
				r1, r2 = self._radius_ratio(ir - 1), self._radius_ratio(ir)
				l0, lN = self._length_ratio(0), self._length_ratio(self.m_nLengthSteps - 1)
				mesh.nAddTriangle(self.vecGetSurfacePoint(lN, p1, r1), self.vecGetSurfacePoint(lN, p1, r2), self.vecGetSurfacePoint(lN, p2, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(lN, p1, r1), self.vecGetSurfacePoint(lN, p2, r2), self.vecGetSurfacePoint(lN, p2, r1))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l0, p1, r1), self.vecGetSurfacePoint(l0, p2, r2), self.vecGetSurfacePoint(l0, p1, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l0, p1, r1), self.vecGetSurfacePoint(l0, p2, r1), self.vecGetSurfacePoint(l0, p2, r2))
			for il in range(1, self.m_nLengthSteps):
				l1, l2 = self._length_ratio(il - 1), self._length_ratio(il)
				r_in = self._radius_ratio(0)
				r_out = self._radius_ratio(self.m_nRadialSteps - 1)
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, p1, r_in), self.vecGetSurfacePoint(l2, p1, r_in), self.vecGetSurfacePoint(l2, p2, r_in))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, p1, r_in), self.vecGetSurfacePoint(l2, p2, r_in), self.vecGetSurfacePoint(l1, p2, r_in))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, p1, r_out), self.vecGetSurfacePoint(l2, p2, r_out), self.vecGetSurfacePoint(l2, p1, r_out))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, p1, r_out), self.vecGetSurfacePoint(l1, p2, r_out), self.vecGetSurfacePoint(l2, p2, r_out))
		return mesh

	def _radius_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nRadialSteps - 1)) * float(step)

	def _phi_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nPolarSteps - 1)) * float(step)

	def _length_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nLengthSteps - 1)) * float(step)

	def vecGetSurfacePoint(self, fLengthRatio: float, fPhiRatio: float, fRadiusRatio: float) -> Vec3:
		spine = as_np3(self.m_aFrames.vecGetSpineAlongLength(fLengthRatio))
		lx = as_np3(self.m_aFrames.vecGetLocalXAlongLength(fLengthRatio))
		phi = (2.0 * math.pi) * fPhiRatio
		outward = self.fGetOuterRadius(fLengthRatio)
		inward = -self.fGetInnerRadius(fLengthRatio)
		radius = fRadiusRatio * (outward - inward) + inward
		pt = spine + radius * lx
		rotated = VecOperations.vecRotateAroundAxis(pt, phi, self.m_oFrame.vecGetLocalZ(), self.m_oFrame.vecGetPosition())
		return self.m_fnTrafo(rotated)

	@staticmethod
	def aGetFramesFromContour(oContour: GenericContour, oFrame: LocalFrame | None = None) -> Frames:
		frame = oFrame if oFrame is not None else LocalFrame()
		samples = 500
		points: list[Vec3] = []
		for i in range(samples):
			lr = float(i) / float(samples - 1)
			z = lr * oContour.m_fTotalLength
			radius = oContour.m_oModulation.fGetModulation(lr)
			rel = (radius, 0.0, z)
			points.append(VecOperations.vecTranslatePointOntoFrame(frame, rel))
		return Frames.from_points_and_frame_type(points, Frames.EFrameType.CYLINDRICAL, 0.5)


__all__ = ["BaseRevolve"]



from __future__ import annotations

import math

import numpy as np

from shape_kernel import LocalFrame, VecOperations
from shape_kernel._types import Vec3, as_np3, as_vec3, normalized

from .icosahedral_face import IcosehedralFace
from .quasi_tile import QuasiTile
from .quasi_tile_01 import QuasiTile_01
from .quasi_tile_03 import QuasiTile_03
from .quasi_tile_04 import QuasiTile_04


class QuasiTileInflation:
    m_oTargetFrame = LocalFrame()
    m_oCurrentFrame = LocalFrame()
    m_oTargetLength = 0.0
    m_oCurrentLength = 1.0

    @staticmethod
    def aGetInflatedFace(sFace: IcosehedralFace) -> list[QuasiTile]:
        vec_face_normal = as_vec3(np.cross(as_np3(sFace.vecLongAxis), as_np3(sFace.vecShortAxis)))
        o_frame = LocalFrame(sFace.vecCentre, vec_face_normal, sFace.vecLongAxis)

        if sFace.eConnector == IcosehedralFace.EConnector.LINE:
            a_inflated_tiles: list[QuasiTile] = []
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt1, sFace.vecPt2, o_frame.vecGetLocalZ(), 0.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt1, sFace.vecPt4, o_frame.vecGetLocalZ(), 1.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt3, sFace.vecPt2, o_frame.vecGetLocalZ(), 1.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt3, sFace.vecPt4, o_frame.vecGetLocalZ(), 0.0 * math.pi / 5.0))
            return a_inflated_tiles
        if sFace.eConnector == IcosehedralFace.EConnector.TRIANGLE:
            a_inflated_tiles = []
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedPurpleLine(sFace.vecPt3, sFace.vecPt4, o_frame.vecGetLocalZ(), 1.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedPurpleLine(sFace.vecPt3, sFace.vecPt2, o_frame.vecGetLocalZ(), 2.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt1, sFace.vecPt4, o_frame.vecGetLocalZ(), 1.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt1, sFace.vecPt2, o_frame.vecGetLocalZ(), 0.0 * math.pi / 5.0))
            return a_inflated_tiles
        if sFace.eConnector == IcosehedralFace.EConnector.ARROW:
            a_inflated_tiles = []
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedPurpleLine(sFace.vecPt3, sFace.vecPt4, o_frame.vecGetLocalZ(), 1.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedPurpleLine(sFace.vecPt1, sFace.vecPt4, o_frame.vecGetLocalZ(), 2.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt3, sFace.vecPt2, o_frame.vecGetLocalZ(), 1.0 * math.pi / 5.0))
            a_inflated_tiles.extend(QuasiTileInflation.aGetInflatedBlackLine(sFace.vecPt1, sFace.vecPt2, o_frame.vecGetLocalZ(), 0.0 * math.pi / 5.0))
            return a_inflated_tiles

        raise Exception("Unknown face connector type. Face cannot get inflated.")

    @staticmethod
    def aGetInflatedBlackLine(vecStart: Vec3, vecEnd: Vec3, vecFaceNormal: Vec3, fCustomAngle: float) -> list[QuasiTile]:
        QuasiTileInflation.m_oTargetLength = float(np.linalg.norm(as_np3(vecEnd) - as_np3(vecStart)))
        vec_target_local_z = as_vec3(normalized(as_np3(vecEnd) - as_np3(vecStart)))
        vec_target_local_x = vecFaceNormal
        QuasiTileInflation.m_oTargetFrame = LocalFrame(vecStart, vec_target_local_z, vec_target_local_x)

        a_inflated_tiles: list[QuasiTile] = []
        o_sub_tile_000 = QuasiTile_01(LocalFrame())
        o_sub_tile_001 = QuasiTile_01(LocalFrame())
        o_sub_tile_002 = QuasiTile_01(LocalFrame())
        o_sub_tile_003 = QuasiTile_01(LocalFrame())
        o_sub_tile_004 = QuasiTile_01(LocalFrame())

        o_sub_tile_001.AttachToOtherQuasiTile(0, o_sub_tile_000, 1)
        o_sub_tile_002.AttachToOtherQuasiTile(0, o_sub_tile_001, 1)
        o_sub_tile_003.AttachToOtherQuasiTile(0, o_sub_tile_002, 1)
        o_sub_tile_004.AttachToOtherQuasiTile(1, o_sub_tile_000, 0)

        o_sub_tile_mid = QuasiTile_03(LocalFrame())
        o_sub_tile_mid.AttachToOtherQuasiTile(17, o_sub_tile_001, 4, True)

        o_sub_tile_005 = QuasiTile_01(LocalFrame())
        o_sub_tile_006 = QuasiTile_01(LocalFrame())
        o_sub_tile_007 = QuasiTile_01(LocalFrame())
        o_sub_tile_008 = QuasiTile_01(LocalFrame())
        o_sub_tile_009 = QuasiTile_01(LocalFrame())

        o_sub_tile_005.AttachToOtherQuasiTile(2, o_sub_tile_mid, 0)
        o_sub_tile_006.AttachToOtherQuasiTile(2, o_sub_tile_mid, 1)
        o_sub_tile_007.AttachToOtherQuasiTile(2, o_sub_tile_mid, 2)
        o_sub_tile_008.AttachToOtherQuasiTile(2, o_sub_tile_mid, 3)
        o_sub_tile_009.AttachToOtherQuasiTile(2, o_sub_tile_mid, 4)

        a_inflated_tiles.extend([
            o_sub_tile_000,
            o_sub_tile_001,
            o_sub_tile_002,
            o_sub_tile_003,
            o_sub_tile_004,
            o_sub_tile_mid,
            o_sub_tile_005,
            o_sub_tile_006,
            o_sub_tile_007,
            o_sub_tile_008,
            o_sub_tile_009,
        ])

        vec_s = o_sub_tile_000.aGetFaces()[2].vecPt3
        vec_e = o_sub_tile_009.aGetFaces()[4].vecPt1
        QuasiTileInflation.m_oCurrentLength = float(np.linalg.norm(as_np3(vec_e) - as_np3(vec_s)))
        vec_current_local_z = as_vec3(normalized(as_np3(vec_e) - as_np3(vec_s)))
        QuasiTileInflation.m_oCurrentFrame = LocalFrame(vec_s, vec_current_local_z)
        vec_ref1 = o_sub_tile_mid.aGetFaces()[-1].vecPt2
        vec_ref2 = o_sub_tile_mid.aGetFaces()[-1].vecPt1
        vec_cross = np.cross(as_np3(vec_current_local_z), normalized(as_np3(vec_ref1) - as_np3(vec_ref2)))
        vec_current_local_x = np.cross(vec_cross, as_np3(vec_current_local_z))
        vec_current_local_x = as_np3(
            VecOperations.vecRotateAroundAxis(
                as_vec3(vec_current_local_x),
                float(fCustomAngle) + math.pi / 10.0,
                vec_current_local_z,
            )
        )
        QuasiTileInflation.m_oCurrentFrame = LocalFrame(vec_s, vec_current_local_z, as_vec3(vec_current_local_x))

        for o_tile in a_inflated_tiles:
            o_tile.ApplyTrafo(QuasiTileInflation.vecTrafo)

        return a_inflated_tiles

    @staticmethod
    def aGetInflatedPurpleLine(vecStart: Vec3, vecEnd: Vec3, vecFaceNormal: Vec3, fCustomAngle: float) -> list[QuasiTile]:
        QuasiTileInflation.m_oTargetLength = float(np.linalg.norm(as_np3(vecEnd) - as_np3(vecStart)))
        vec_target_local_z = as_vec3(normalized(as_np3(vecEnd) - as_np3(vecStart)))
        vec_target_local_x = vecFaceNormal
        QuasiTileInflation.m_oTargetFrame = LocalFrame(vecStart, vec_target_local_z, vec_target_local_x)

        a_inflated_tiles: list[QuasiTile] = []
        o_sub_tile_000 = QuasiTile_01(LocalFrame())
        o_sub_tile_001 = QuasiTile_01(LocalFrame())
        o_sub_tile_002 = QuasiTile_01(LocalFrame())
        o_sub_tile_003 = QuasiTile_01(LocalFrame())
        o_sub_tile_004 = QuasiTile_01(LocalFrame())

        o_sub_tile_001.AttachToOtherQuasiTile(0, o_sub_tile_000, 1)
        o_sub_tile_002.AttachToOtherQuasiTile(0, o_sub_tile_001, 1)
        o_sub_tile_003.AttachToOtherQuasiTile(0, o_sub_tile_002, 1)
        o_sub_tile_004.AttachToOtherQuasiTile(1, o_sub_tile_000, 0)

        o_sub_tile_mid = QuasiTile_04(LocalFrame())
        o_sub_tile_mid.AttachToOtherQuasiTile(17, o_sub_tile_001, 4, True)

        a_inflated_tiles.extend([
            o_sub_tile_000,
            o_sub_tile_001,
            o_sub_tile_002,
            o_sub_tile_003,
            o_sub_tile_004,
            o_sub_tile_mid,
        ])

        vec_s = o_sub_tile_000.aGetFaces()[2].vecPt3
        vec_e = o_sub_tile_mid.aGetFaces()[4].vecPt1
        QuasiTileInflation.m_oCurrentLength = float(np.linalg.norm(as_np3(vec_e) - as_np3(vec_s)))
        vec_current_local_z = as_vec3(normalized(as_np3(vec_e) - as_np3(vec_s)))
        QuasiTileInflation.m_oCurrentFrame = LocalFrame(vec_s, vec_current_local_z)
        vec_ref1 = o_sub_tile_mid.aGetFaces()[2].vecPt2
        vec_ref2 = o_sub_tile_mid.aGetFaces()[2].vecPt1
        vec_cross = np.cross(as_np3(vec_current_local_z), normalized(as_np3(vec_ref1) - as_np3(vec_ref2)))
        vec_current_local_x = np.cross(vec_cross, as_np3(vec_current_local_z))
        vec_current_local_x = as_np3(
            VecOperations.vecRotateAroundAxis(
                as_vec3(vec_current_local_x),
                float(fCustomAngle) + math.pi / 10.0,
                vec_current_local_z,
            )
        )
        QuasiTileInflation.m_oCurrentFrame = LocalFrame(vec_s, vec_current_local_z, as_vec3(vec_current_local_x))

        for o_tile in a_inflated_tiles:
            o_tile.ApplyTrafo(QuasiTileInflation.vecTrafo)

        return a_inflated_tiles

    @staticmethod
    def vecTrafo(vecPt: Vec3) -> Vec3:
        vec_rel = as_np3(VecOperations.vecExpressPointInFrame(QuasiTileInflation.m_oCurrentFrame, vecPt))
        vec_rel *= QuasiTileInflation.m_oTargetLength / QuasiTileInflation.m_oCurrentLength
        vec_new_pt = VecOperations.vecTranslatePointOntoFrame(QuasiTileInflation.m_oTargetFrame, as_vec3(vec_rel))
        return vec_new_pt


__all__ = ["QuasiTileInflation"]



from __future__ import annotations

from typing import Callable

from picogk import Image

from .._types import Vec3, as_np3, as_vec3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from ..modulations import LineModulation, SurfaceModulation
from .base_box import BaseBox
from .base_shape import BaseShape


class BaseLogoBox(BaseBox):
	def __init__(self, oFrame: LocalFrame, fLength: float, fRefWidth: float, oImage: Image, oMappingFunc: Callable[[float], float]) -> None:
		BaseShape.__init__(self)
		self.m_aFrames = Frames.from_length(fLength, oFrame)
		self.m_nWidthSteps = int(oImage.nWidth)
		self.m_nDepthSteps = int(oImage.nHeight)
		self.m_nLengthSteps = 5
		fDepth = (float(oImage.nHeight) / float(oImage.nWidth)) * fRefWidth
		self.m_oWidthModulation = LineModulation(fRefWidth)
		self.m_oDepthModulation = LineModulation(fDepth)
		self.m_oTopModulation = SurfaceModulation.from_image(oImage, oMappingFunc)

	def fGetWidth(self) -> float:
		return self.m_oWidthModulation.m_fConstValue

	def fGetDepth(self) -> float:
		return self.m_oDepthModulation.m_fConstValue

	def vecGetSurfacePoint(self, fWidthRatio: float, fDepthRatio: float, fLengthRatio: float) -> Vec3:
		spine = as_np3(self.m_aFrames.vecGetSpineAlongLength(fLengthRatio))
		lx = as_np3(self.m_aFrames.vecGetLocalXAlongLength(fLengthRatio))
		ly = as_np3(self.m_aFrames.vecGetLocalYAlongLength(fLengthRatio))
		lz = as_np3(self.m_aFrames.vecGetLocalZAlongLength(fLengthRatio))

		x = 0.5 * fWidthRatio * self.m_oWidthModulation.m_fConstValue
		y = 0.5 * fDepthRatio * self.m_oDepthModulation.m_fConstValue
		pt = spine + x * lx + y * ly

		if abs(fLengthRatio - 1.0) < 0.0003:
			img_w = 1.0 - (0.5 + 0.5 * fWidthRatio)
			img_h = 1.0 - (0.5 + 0.5 * fDepthRatio)
			dz = self.m_oTopModulation.fGetModulation(img_w, img_h)
			pt = pt + dz * lz

		return self.m_fnTrafo(as_vec3(pt))


__all__ = ["BaseLogoBox"]



from abc import ABC, abstractmethod

from .._types import Vec3

class ISpline(ABC):
    @abstractmethod
    def aGetPoints(self, nSamples: int | None = 500) -> list[Vec3]:
        raise NotImplementedError




from __future__ import annotations

import math
from enum import Enum

import numpy as np

from picogk import Lattice
from shape_kernel import Cp, LocalFrame, Sh, VecOperations
from shape_kernel._types import as_np3, as_vec3, normalized


class IcosehedralFace:
    class EConnector(Enum):
        ARROW = "ARROW"
        TRIANGLE = "TRIANGLE"
        LINE = "LINE"

    class EDef(Enum):
        CENTRE = "CENTRE"
        SHORT_AXIS = "SHORT_AXIS"
        LONG_AXIS = "LONG_AXIS"

    m_fPsi = math.acos(1.0 / math.sqrt(5.0))

    def __init__(
        self,
        oFrame: LocalFrame,
        eDef: "IcosehedralFace.EDef",
        eConnector: "IcosehedralFace.EConnector",
        fSide: float = 20.0,
    ) -> None:
        self.eConnector = eConnector

        vec_pointer = float(fSide) * np.array((1.0, 0.0, 0.0), dtype=np.float64)
        vec_pointer_01 = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(vec_pointer), -0.5 * self.m_fPsi, (0.0, 0.0, 1.0)))
        vec_pointer_02 = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(vec_pointer), +0.5 * self.m_fPsi, (0.0, 0.0, 1.0)))

        self.vecPt1 = as_vec3((0.0, 0.0, 0.0))
        self.vecPt2 = as_vec3(as_np3(self.vecPt1) + vec_pointer_01)
        self.vecPt3 = as_vec3(as_np3(self.vecPt2) + vec_pointer_02)
        self.vecPt4 = as_vec3(as_np3(self.vecPt3) - vec_pointer_01)

        if eDef == self.EDef.CENTRE:
            vec_centre = as_np3(self.vecPt1) + 0.5 * (as_np3(self.vecPt3) - as_np3(self.vecPt1))
            self.vecPt1 = as_vec3(as_np3(self.vecPt1) - vec_centre)
            self.vecPt2 = as_vec3(as_np3(self.vecPt2) - vec_centre)
            self.vecPt3 = as_vec3(as_np3(self.vecPt3) - vec_centre)
            self.vecPt4 = as_vec3(as_np3(self.vecPt4) - vec_centre)

            self.vecPt1 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt1)
            self.vecPt2 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt2)
            self.vecPt3 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt3)
            self.vecPt4 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt4)
        elif eDef == self.EDef.LONG_AXIS:
            vec_shift = as_np3(self.vecPt1)
            self.vecPt1 = as_vec3(as_np3(self.vecPt1) - vec_shift)
            self.vecPt2 = as_vec3(as_np3(self.vecPt2) - vec_shift)
            self.vecPt3 = as_vec3(as_np3(self.vecPt3) - vec_shift)
            self.vecPt4 = as_vec3(as_np3(self.vecPt4) - vec_shift)

            self.vecPt1 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt1)
            self.vecPt2 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt2)
            self.vecPt3 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt3)
            self.vecPt4 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt4)
        elif eDef == self.EDef.SHORT_AXIS:
            vec_centre = as_np3(self.vecPt1) + 0.5 * (as_np3(self.vecPt3) - as_np3(self.vecPt1))
            self.vecPt1 = as_vec3(as_np3(self.vecPt1) - vec_centre)
            self.vecPt2 = as_vec3(as_np3(self.vecPt2) - vec_centre)
            self.vecPt3 = as_vec3(as_np3(self.vecPt3) - vec_centre)
            self.vecPt4 = as_vec3(as_np3(self.vecPt4) - vec_centre)

            self.vecPt1 = VecOperations.vecRotateAroundZ(self.vecPt1, -90.0 / 180.0 * math.pi)
            self.vecPt2 = VecOperations.vecRotateAroundZ(self.vecPt2, -90.0 / 180.0 * math.pi)
            self.vecPt3 = VecOperations.vecRotateAroundZ(self.vecPt3, -90.0 / 180.0 * math.pi)
            self.vecPt4 = VecOperations.vecRotateAroundZ(self.vecPt4, -90.0 / 180.0 * math.pi)

            vec_shift = as_np3(self.vecPt2)
            self.vecPt1 = as_vec3(as_np3(self.vecPt1) - vec_shift)
            self.vecPt2 = as_vec3(as_np3(self.vecPt2) - vec_shift)
            self.vecPt3 = as_vec3(as_np3(self.vecPt3) - vec_shift)
            self.vecPt4 = as_vec3(as_np3(self.vecPt4) - vec_shift)

            self.vecPt1 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt1)
            self.vecPt2 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt2)
            self.vecPt3 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt3)
            self.vecPt4 = VecOperations.vecTranslatePointOntoFrame(oFrame, self.vecPt4)

        self.vecCentre = as_vec3(as_np3(self.vecPt1) + 0.5 * (as_np3(self.vecPt3) - as_np3(self.vecPt1)))
        self.vecLongAxis = as_vec3(normalized(as_np3(self.vecPt3) - as_np3(self.vecPt1)))
        self.vecShortAxis = as_vec3(normalized(as_np3(self.vecPt4) - as_np3(self.vecPt2)))

    def FlipAroundShortAxis(self) -> None:
        vec_temp1 = self.vecPt1
        vec_temp2 = self.vecPt2
        vec_temp3 = self.vecPt3
        vec_temp4 = self.vecPt4

        self.vecPt1 = vec_temp3
        self.vecPt2 = vec_temp2
        self.vecPt3 = vec_temp1
        self.vecPt4 = vec_temp4

    def FlipAroundLongAxis(self) -> None:
        vec_temp1 = self.vecPt1
        vec_temp2 = self.vecPt2
        vec_temp3 = self.vecPt3
        vec_temp4 = self.vecPt4

        self.vecPt1 = vec_temp1
        self.vecPt2 = vec_temp4
        self.vecPt3 = vec_temp3
        self.vecPt4 = vec_temp2

    def Preview(self, bShowConnector: bool) -> None:
        a_boundary = [self.vecPt1, self.vecPt2, self.vecPt3, self.vecPt4, self.vecPt1]
        a_long_axis = [self.vecPt1, self.vecPt3]
        a_short_axis = [self.vecPt2, self.vecPt4]
        Sh.PreviewLine(a_boundary, Cp.clrBlack)
        Sh.PreviewLine(a_long_axis, Cp.clrGreen)
        Sh.PreviewLine(a_short_axis, Cp.clrRed)

        if bShowConnector:
            lat_connector = Lattice()
            if self.eConnector == self.EConnector.LINE:
                vec_start = as_vec3(as_np3(self.vecPt1) + 0.2 * (as_np3(self.vecPt3) - as_np3(self.vecPt1)))
                vec_end = as_vec3(as_np3(self.vecPt1) + 0.8 * (as_np3(self.vecPt3) - as_np3(self.vecPt1)))
                lat_connector.AddBeam(vec_start, vec_end, 0.5, 0.5, True)
                Sh.PreviewLattice(lat_connector, Cp.clrBlue)
            elif self.eConnector == self.EConnector.ARROW:
                vec_start = as_vec3(as_np3(self.vecPt2) + 0.2 * (as_np3(self.vecPt4) - as_np3(self.vecPt2)))
                vec_end = as_vec3(as_np3(self.vecPt2) + 0.8 * (as_np3(self.vecPt4) - as_np3(self.vecPt2)))
                lat_connector.AddBeam(vec_start, vec_end, 0.5, 0.5, True)

                vec_arrow1 = as_vec3(
                    as_np3(self.vecPt2)
                    + 0.4 * (as_np3(self.vecPt4) - as_np3(self.vecPt2))
                    + 0.2 * (as_np3(self.vecPt3) - as_np3(self.vecPt1))
                )
                vec_arrow2 = as_vec3(
                    as_np3(self.vecPt2)
                    + 0.4 * (as_np3(self.vecPt4) - as_np3(self.vecPt2))
                    - 0.2 * (as_np3(self.vecPt3) - as_np3(self.vecPt1))
                )
                lat_connector.AddBeam(vec_start, vec_arrow1, 0.5, 0.5, True)
                lat_connector.AddBeam(vec_start, vec_arrow2, 0.5, 0.5, True)
                Sh.PreviewLattice(lat_connector, Cp.clrRed)
            elif self.eConnector == self.EConnector.TRIANGLE:
                vec_start = as_vec3(as_np3(self.vecPt2) + 0.2 * (as_np3(self.vecPt4) - as_np3(self.vecPt2)))
                vec_end = as_vec3(as_np3(self.vecPt2) + 0.8 * (as_np3(self.vecPt4) - as_np3(self.vecPt2)))
                lat_connector.AddBeam(vec_start, vec_end, 0.5, 0.5, True)

                vec_tri = as_vec3(
                    as_np3(self.vecPt2)
                    + 0.5 * (as_np3(self.vecPt4) - as_np3(self.vecPt2))
                    + 0.25 * (as_np3(self.vecPt3) - as_np3(self.vecPt1))
                )
                lat_connector.AddBeam(vec_start, vec_tri, 0.5, 0.5, True)
                lat_connector.AddBeam(vec_end, vec_tri, 0.5, 0.5, True)
                Sh.PreviewLattice(lat_connector, Cp.clrBillie)


__all__ = ["IcosehedralFace"]



from __future__ import annotations

import math

import numpy as np

from shape_kernel import Cp, LocalFrame, VecOperations
from shape_kernel._types import as_np3, as_vec3, normalized

from .icosahedral_face import IcosehedralFace
from .quasi_tile import QuasiTile


class QuasiTile_03(QuasiTile):
    def __init__(self, oFrame0c: LocalFrame, fFaceSide: float = 20.0) -> None:
        super().__init__()
        self.m_clr = Cp.clrCrystal

        f_rot_angle = (360.0 / 5.0) / 180.0 * math.pi

        s_ref_face = IcosehedralFace(oFrame0c, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.LINE, fFaceSide)
        f_long_axis = float(np.linalg.norm(as_np3(s_ref_face.vecPt3) - as_np3(s_ref_face.vecPt1)))
        f_short_axis = float(np.linalg.norm(as_np3(s_ref_face.vecPt4) - as_np3(s_ref_face.vecPt2)))
        f_dome_pentagon_side = f_short_axis
        f_dome_pentagon_height = f_dome_pentagon_side / (2.0 * math.sqrt(5.0 - math.sqrt(20.0)))
        f_tilt_angle_rad = math.asin(f_dome_pentagon_height / (0.5 * f_long_axis))
        f_tilt_angle_deg = 90.0 - (f_tilt_angle_rad / math.pi * 180.0)
        f_tilt_angle = (-f_tilt_angle_deg) / 180.0 * math.pi

        a_lower_centre_faces: list[IcosehedralFace] = []
        for i in range(5):
            o_frame_1b = LocalFrame.oGetRotatedFrame(oFrame0c, i * f_rot_angle, oFrame0c.vecGetLocalZ())
            o_frame_1b = LocalFrame.oGetRotatedFrame(o_frame_1b, f_tilt_angle, o_frame_1b.vecGetLocalY())

            s_face_1b = IcosehedralFace(o_frame_1b, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.TRIANGLE, fFaceSide)
            a_lower_centre_faces.append(s_face_1b)

        a_lower_side_faces: list[IcosehedralFace] = []
        for i in range(5):
            i_lower_index = i - 1
            if i_lower_index < 0:
                i_lower_index += 5
            i_upper_index = i

            s_face_1b = a_lower_centre_faces[i_lower_index]
            s_face_2b = a_lower_centre_faces[i_upper_index]

            vec_tip_1b = s_face_1b.vecPt3
            vec_tip_2b = s_face_2b.vecPt3
            vec_long_1s = as_vec3(normalized(as_np3(vec_tip_2b) - as_np3(vec_tip_1b)))
            vec_centre_1s = as_vec3(as_np3(vec_tip_1b) + 0.5 * (as_np3(vec_tip_2b) - as_np3(vec_tip_1b)))
            vec_short_1s = as_vec3(normalized(as_np3(s_face_1b.vecPt4) - as_np3(vec_centre_1s)))
            vec_normal_1s = as_vec3(np.cross(as_np3(vec_long_1s), as_np3(vec_short_1s)))
            o_frame_1s = LocalFrame(vec_tip_1b, vec_normal_1s, vec_long_1s)

            s_face_1s = IcosehedralFace(o_frame_1s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.ARROW, fFaceSide)
            a_lower_side_faces.append(s_face_1s)

        f_lower_z = VecOperations.vecExpressPointInFrame(oFrame0c, a_lower_side_faces[0].vecPt1)[2]
        f_upper_z = VecOperations.vecExpressPointInFrame(oFrame0c, a_lower_side_faces[0].vecPt2)[2]
        f_max_z = f_lower_z + f_upper_z
        o_frame_1c = LocalFrame.oGetTranslatedFrame(oFrame0c, tuple(f_max_z * x for x in oFrame0c.vecGetLocalZ()))
        o_frame_1c = LocalFrame.oGetInvertFrame(o_frame_1c, True, True)

        a_upper_centre_faces: list[IcosehedralFace] = []
        for i in range(5):
            o_frame_1t = LocalFrame.oGetRotatedFrame(o_frame_1c, i * f_rot_angle, o_frame_1c.vecGetLocalZ())
            o_frame_1t = LocalFrame.oGetRotatedFrame(o_frame_1t, f_tilt_angle, o_frame_1t.vecGetLocalY())

            s_face_1t = IcosehedralFace(o_frame_1t, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.LINE, fFaceSide)
            a_upper_centre_faces.append(s_face_1t)

        a_upper_side_faces: list[IcosehedralFace] = []
        for i in range(5):
            i_lower_index = i - 1
            if i_lower_index < 0:
                i_lower_index += 5
            i_upper_index = i

            s_face_1b = a_upper_centre_faces[i_lower_index]
            s_face_2b = a_upper_centre_faces[i_upper_index]

            vec_tip_1b = s_face_1b.vecPt3
            vec_tip_2b = s_face_2b.vecPt3
            vec_long_1s = as_vec3(normalized(as_np3(vec_tip_2b) - as_np3(vec_tip_1b)))
            vec_centre_1s = as_vec3(as_np3(vec_tip_1b) + 0.5 * (as_np3(vec_tip_2b) - as_np3(vec_tip_1b)))
            vec_short_1s = as_vec3(normalized(as_np3(s_face_1b.vecPt4) - as_np3(vec_centre_1s)))
            vec_normal_1s = as_vec3(np.cross(as_np3(vec_long_1s), as_np3(vec_short_1s)))
            o_frame_1s = LocalFrame(vec_tip_1b, vec_normal_1s, vec_long_1s)

            s_face_1s = IcosehedralFace(o_frame_1s, IcosehedralFace.EDef.LONG_AXIS, IcosehedralFace.EConnector.LINE, fFaceSide)
            a_upper_side_faces.append(s_face_1s)

        self.m_aFaces = []
        self.m_aFaces.extend(a_lower_centre_faces)
        self.m_aFaces.extend(a_lower_side_faces)
        self.m_aFaces.extend(a_upper_side_faces)
        self.m_aFaces.extend(a_upper_centre_faces)


__all__ = ["QuasiTile_03"]



from .base_shapes import (
    BaseBox,
    BaseCone,
    BaseCylinder,
    BaseLens,
    BaseLogoBox,
    BasePipe,
    BasePipeSegment,
    BaseRevolve,
    BaseRing,
    BaseShape,
    BaseSphere,
    LatticeManifold,
    LatticePipe,
)
from .frames import Frames, LocalFrame
from .modulations import Distribution, GenericContour, LineModulation, SurfaceModulation
from .splines import (
    ControlPointSpline,
    ControlPointSurface,
    CylindricalControlSpline,
    # ISpline,
    TangentialControlSpline,
)
from .utils import (
    Bisection,
    BisectionException,
    CSVWriter,
    GridOperations,
    ImplicitGenus,
    ImplicitGyroid,
    ImplicitSphere,
    ImplicitSuperEllipsoid,
    LineDecimation,
    ListOperations,
    Uf,
    VecOperations,
    SplineOperations,
)
from .functions import Sh
from .utils import CylUtility, Measure, MeshUtility
from .visualizations import (
    Cp,
    ColorScale3D,
    CustomColorScale2D,
    IColorScale,
    ISpectrum,
    LinearColorScale2D,
    MeshPainter,
    RainboxSpectrum,
    RotationAnimator,
    SmoothColorScale2D,
)

__all__ = [
    # "ISpline",
    "LocalFrame",
    "VecOperations",
    "Frames",
    "LineModulation",
    "SurfaceModulation",
    "Distribution",
    "GenericContour",
    "ControlPointSpline",
    "CylindricalControlSpline",
    "ControlPointSurface",
    "TangentialControlSpline",
    "SplineOperations",
    "Bisection",
    "BisectionException",
    "ImplicitGyroid",
    "ImplicitSphere",
    "ImplicitGenus",
    "ImplicitSuperEllipsoid",
    "GridOperations",
    "Sh",
    "IColorScale",
    "LinearColorScale2D",
    "SmoothColorScale2D",
    "CustomColorScale2D",
    "ISpectrum",
    "RainboxSpectrum",
    "ColorScale3D",
    "MeshPainter",
    "RotationAnimator",
    "Cp",
    "BaseShape",
    "BaseBox",
    "BaseCone",
    "BaseCylinder",
    "BaseLens",
    "BaseLogoBox",
    "BasePipe",
    "BasePipeSegment",
    "BaseRevolve",
    "BaseRing",
    "BaseSphere",
    "LatticePipe",
    "LatticeManifold",
    "ListOperations",
    "LineDecimation",
    "CSVWriter",
    "Uf",
    "Measure",
    "MeshUtility",
    "CylUtility",
]



from __future__ import annotations

import math
from abc import ABC, abstractmethod
from collections.abc import Sequence


class IColorScale(ABC):
	@abstractmethod
	def clrGetColor(self, fValue: float) -> tuple[float, float, float, float]:
		raise NotImplementedError

	@abstractmethod
	def fGetMinValue(self) -> float:
		raise NotImplementedError

	@abstractmethod
	def fGetMaxValue(self) -> float:
		raise NotImplementedError


class LinearColorScale2D(IColorScale):
	def __init__(
		self,
		clrMin: Sequence[float],
		clrMax: Sequence[float],
		fMinValue: float,
		fMaxValue: float,
	) -> None:
		self.m_fMinValue = float(fMinValue)
		self.m_fMaxValue = float(fMaxValue)
		self.m_clrMin = tuple(float(v) for v in clrMin[:3])
		self.m_clrMax = tuple(float(v) for v in clrMax[:3])

	def fGetMinValue(self) -> float:
		return self.m_fMinValue

	def fGetMaxValue(self) -> float:
		return self.m_fMaxValue

	def fGetInterpolated(self, fValue1: float, fValue2: float, fValue: float) -> float:
		ratio = (fValue - self.m_fMinValue) / (self.m_fMaxValue - self.m_fMinValue)
		return fValue1 + ratio * (fValue2 - fValue1)

	def clrGetColor(self, fValue: float) -> tuple[float, float, float, float]:
		value = min(max(float(fValue), self.m_fMinValue), self.m_fMaxValue)
		r = self.fGetInterpolated(self.m_clrMin[0], self.m_clrMax[0], value)
		g = self.fGetInterpolated(self.m_clrMin[1], self.m_clrMax[1], value)
		b = self.fGetInterpolated(self.m_clrMin[2], self.m_clrMax[2], value)
		return (float(r), float(g), float(b), 1.0)


class SmoothColorScale2D(LinearColorScale2D):
	def fGetInterpolated(self, fValue1: float, fValue2: float, fValue: float) -> float:
		ratio = (fValue - self.m_fMinValue) / (self.m_fMaxValue - self.m_fMinValue)
		ratio = ratio * ratio * (3.0 - 2.0 * ratio)
		return fValue1 + ratio * (fValue2 - fValue1)


class CustomColorScale2D(LinearColorScale2D):
	def __init__(
		self,
		clrMin: Sequence[float],
		clrMax: Sequence[float],
		fMinValue: float,
		fMaxValue: float,
		fTransition: float,
		fSmoothness: float,
	) -> None:
		super().__init__(clrMin, clrMax, fMinValue, fMaxValue)
		self.m_fTransition = float(fTransition)
		self.m_fSmoothness = float(fSmoothness)

	def fGetInterpolated(self, fValue1: float, fValue2: float, fValue: float) -> float:
		x = float(fValue)
		s = max(1e-9, self.m_fSmoothness)
		t = self.m_fTransition
		ratio = 1.0 / (1.0 + math.exp(-(x - t) / s))
		return fValue1 + ratio * (fValue2 - fValue1)


__all__ = ["IColorScale", "LinearColorScale2D", "SmoothColorScale2D", "CustomColorScale2D"]



from _common.types import ColorLike, clamp01

def to_rgba(color: ColorLike) -> tuple[float, float, float, float]:
    if len(color) == 4:
        return (clamp01(color[0]), clamp01(color[1]), clamp01(color[2]), clamp01(color[3]))
    if len(color) == 3:
        return (clamp01(color[0]), clamp01(color[1]), clamp01(color[2]), 1.0)
    if len(color) == 1:
        c = clamp01(color[0])
        return (c, c, c, 1.0)
    raise ValueError("Color must have 1, 3, or 4 channels")






from __future__ import annotations

from picogk import Lattice
from shape_kernel import SplineOperations, Uf
from shape_kernel._types import as_np3, as_vec3

from .._helpers import _add_polyline_beam
from ..beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell
from .i_lattice_type import ILatticeType


class RandomSplineLattice(ILatticeType):
	def __init__(self, nPasses: int = 1) -> None:
		self.m_nPasses = int(nPasses)
		self.m_dX = 0.0
		self.m_dY = 0.0
		self.m_dZ = 0.0

	def AddCell(self, oLattice: Lattice, xCell: IUnitCell, xBeamThickness: IBeamThickness, nSubSamples: int = 2) -> None:
		vecSize = xCell.oGetCellBounding().vecSize()
		self.m_dX = float(vecSize[0])
		self.m_dY = float(vecSize[1])
		self.m_dZ = float(vecSize[2])

		aCornerPoints = xCell.aGetCornerPoints()
		for _ in range(self.m_nPasses):
			for i, vecStart in enumerate(aCornerPoints):
				j = i
				while j == i:
					j = int(Uf.fGetRandomLinear(0.0, float(len(aCornerPoints))))

				aPoints = [
					as_vec3(vecStart),
					as_vec3(as_np3(xCell.vecGetCellCentre()) + as_np3(self.vecGetNoise())),
					as_vec3(aCornerPoints[j]),
				]
				aSplinePoints = SplineOperations.aGetNURBSpline(aPoints, 20)
				_add_polyline_beam(oLattice, aSplinePoints, xBeamThickness)

	def vecGetNoise(self) -> tuple[float, float, float]:
		return (
			Uf.fGetRandomLinear(-0.3 * self.m_dX, 0.3 * self.m_dX),
			Uf.fGetRandomLinear(-0.3 * self.m_dY, 0.3 * self.m_dY),
			Uf.fGetRandomLinear(-0.3 * self.m_dZ, 0.3 * self.m_dZ),
		)


__all__ = ["RandomSplineLattice"]


from __future__ import annotations

import ctypes
from enum import IntEnum
from typing import Any, cast

from .._base import HandleOwner
from .._config import STRING_LENGTH
from .._errors import PicoGKInvalidHandleError
from .fields import ScalarField, VectorField
from .library import Library
from .._core.voxels import Voxels

VdbField = Voxels | ScalarField | VectorField

class OpenVdbFile(HandleOwner):
    class FieldType(IntEnum):
        UNSUPPORTED = -1
        VOXELS = 0
        SCALAR_FIELD = 1
        VECTOR_FIELD = 2

    def __init__(self, path: str | None = None, handle: int | ctypes.c_void_p | None = None, *, owns_handle: bool = True) -> None:
        if handle is None and path is None:
            handle = Library._lib().VdbFile_hCreate()
        elif handle is None and path is not None:
            handle = Library._lib().VdbFile_hCreateFromFile(path.encode("utf-8"))
        if not handle:
            raise PicoGKInvalidHandleError("VdbFile creation returned null")
        super().__init__(handle, owns_handle=owns_handle)

    def _destroy_native(self) -> None:
        Library._lib().VdbFile_Destroy(self._handle)

    def save(self, path: str) -> None:
        self._ensure_open()
        for index in range(self.field_count()):
            field = cast(Any, self.xField(index))
            field.oMetaData().SetValue("PicoGK.Library", Library.name())
            field.oMetaData().SetValue("PicoGK.Version", Library.version())
            field.oMetaData().SetValue("PicoGK.VoxelSize", Library.voxel_size_mm / 1000.0)
        ok = bool(Library._lib().VdbFile_bSaveToFile(self.handle, path.encode("utf-8")))
        if not ok:
            raise IOError(f"Failed to save VDB: {path}")

    SaveToFile = save

    def field_count(self) -> int:
        self._ensure_open()
        return int(Library._lib().VdbFile_nFieldCount(self.handle))

    nFieldCount = field_count

    def field_name(self, index: int) -> str:
        self._ensure_open()
        buf = ctypes.create_string_buffer(STRING_LENGTH)
        Library._lib().VdbFile_GetFieldName(self.handle, int(index), buf)
        return buf.value.decode("utf-8")

    strFieldName = field_name

    def field_type(self, index: int) -> "OpenVdbFile.FieldType":
        self._ensure_open()
        return OpenVdbFile.FieldType(int(Library._lib().VdbFile_nFieldType(self.handle, int(index))))

    eFieldType = field_type

    def strFieldType(self, index: int) -> str:
        mapping = {
            self.FieldType.UNSUPPORTED: "Unsupported",
            self.FieldType.VOXELS: "Voxels",
            self.FieldType.SCALAR_FIELD: "ScalarField",
            self.FieldType.VECTOR_FIELD: "VectorField",
        }
        return mapping.get(self.field_type(index), f"Unknown #{self.field_type(index).value}")

    def add_voxels(self, voxels: Voxels, field_name: str = "") -> int:
        self._ensure_open()
        if not field_name:
            field_name = f"PicoGK.Voxels.{self.field_count()}"
        return int(Library._lib().VdbFile_nAddVoxels(self.handle, field_name.encode("utf-8"), voxels.handle))

    def add_scalar_field(self, field: ScalarField, field_name: str = "") -> int:
        self._ensure_open()
        if not field_name:
            field_name = f"PicoGK.ScalarField.{self.field_count()}"
        return int(Library._lib().VdbFile_nAddScalarField(self.handle, field_name.encode("utf-8"), field.handle))

    def add_vector_field(self, field: VectorField, field_name: str = "") -> int:
        self._ensure_open()
        if not field_name:
            field_name = f"PicoGK.VectorField.{self.field_count()}"
        return int(Library._lib().VdbFile_nAddVectorField(self.handle, field_name.encode("utf-8"), field.handle))

    def nAdd(self, field: VdbField, field_name: str = "") -> int:
        if isinstance(field, Voxels):
            return self.add_voxels(field, field_name)
        if isinstance(field, ScalarField):
            return self.add_scalar_field(field, field_name)
        if isinstance(field, VectorField):
            return self.add_vector_field(field, field_name)
        raise TypeError("nAdd expects Voxels, ScalarField, or VectorField")

    def get_voxels(self, index: int) -> Voxels:
        self._ensure_open()
        h = Library._lib().VdbFile_hGetVoxels(self.handle, int(index))
        if not h:
            raise PicoGKInvalidHandleError(f"No voxel field at index {index}")
        return Voxels(handle=h)

    def voxGet(self, index_or_name: int | str) -> Voxels:
        if isinstance(index_or_name, int):
            return self.get_voxels(index_or_name)
        for index in range(self.field_count()):
            if self.field_name(index).lower() == index_or_name.lower():
                return self.get_voxels(index)
        raise KeyError(index_or_name)

    def get_scalar_field(self, index: int) -> ScalarField:
        self._ensure_open()
        h = Library._lib().VdbFile_hGetScalarField(self.handle, int(index))
        if not h:
            raise PicoGKInvalidHandleError(f"No scalar field at index {index}")
        return ScalarField(handle=h)

    def oGetScalarField(self, index_or_name: int | str) -> ScalarField:
        if isinstance(index_or_name, int):
            return self.get_scalar_field(index_or_name)
        for index in range(self.field_count()):
            if self.field_name(index).lower() == index_or_name.lower():
                return self.get_scalar_field(index)
        raise KeyError(index_or_name)

    def get_vector_field(self, index: int) -> VectorField:
        self._ensure_open()
        h = Library._lib().VdbFile_hGetVectorField(self.handle, int(index))
        if not h:
            raise PicoGKInvalidHandleError(f"No vector field at index {index}")
        return VectorField(handle=h)

    def oGetVectorField(self, index_or_name: int | str) -> VectorField:
        if isinstance(index_or_name, int):
            return self.get_vector_field(index_or_name)
        for index in range(self.field_count()):
            if self.field_name(index).lower() == index_or_name.lower():
                return self.get_vector_field(index)
        raise KeyError(index_or_name)

    def xField(self, index: int) -> VdbField:
        field_type = self.field_type(index)
        if field_type == self.FieldType.VOXELS:
            return self.get_voxels(index)
        if field_type == self.FieldType.SCALAR_FIELD:
            return self.get_scalar_field(index)
        if field_type == self.FieldType.VECTOR_FIELD:
            return self.get_vector_field(index)
        raise ValueError(f"Unsupported field at index {index}")

    def bIsPicoGKCompatible(self) -> bool:
        if self.field_count() < 1:
            return False
        field = cast(Any, self.xField(0))
        ok, _ = field.oMetaData().bGetValueAt("PicoGK.VoxelSize")
        return ok

    def fPicoGKVoxelSizeMM(self) -> float:
        if self.field_count() < 1:
            return 0.0
        field = cast(Any, self.xField(0))
        ok, value = field.oMetaData().bGetValueAt("PicoGK.VoxelSize")
        if ok and isinstance(value, (float, int)) and float(value) > 0:
            return float(value) * 1000.0
        return 0.0






from __future__ import annotations

from enum import IntEnum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .fields import ScalarField

import ctypes
from typing import Any, Sequence, cast

import numpy as np

from _common.types import Vector3Like

from .._base import HandleOwner
from .._errors import PicoGKInvalidHandleError
from .._native import CallbackImplicitDistance
from .._types import BBox3, Vec3, VoxelDimensions, as_bbox3, as_vec3, vec3_tuple

from .lattice import Lattice
from .library import Library
from .mesh import Mesh
from .metadata import FieldMetadata

from abc import ABC, abstractmethod


class IImplicit(ABC):
	@abstractmethod
	def fSignedDistance(self, vecPt: Vector3Like) -> float:
		raise NotImplementedError


class IBoundedImplicit(IImplicit):
    @property
    @abstractmethod
    def oBounds(self) -> tuple[Vector3Like, Vector3Like]:
        raise NotImplementedError


__all__ = ["IImplicit", "IBoundedImplicit", "Voxels"]



def _create_native_distance_callback(implicit: IImplicit) -> Any:
    def _wrapped(vec_ptr: Any) -> float:
        v = vec_ptr.contents
        return float(implicit.fSignedDistance((float(v.x), float(v.y), float(v.z))))

    return CallbackImplicitDistance(_wrapped)

class ESliceMode(IntEnum):
    SignedDistance = 0
    Antialiased = 1
    BlackWhite = 2

class Voxels(HandleOwner):
    def __init__(self, handle: int | ctypes.c_void_p | None = None, *, copy_from: "Voxels | None" = None, owns_handle: bool = True) -> None:
        if handle is None and copy_from is not None:
            handle = Library._lib().Voxels_hCreateCopy(copy_from.handle)
        elif handle is None:
            handle = Library._lib().Voxels_hCreate()
        if not handle:
            raise PicoGKInvalidHandleError("Voxels handle creation returned null")
        super().__init__(handle, owns_handle=owns_handle)
        self.metadata = FieldMetadata.from_voxels(self)
        self.metadata.set_string("PicoGK.Class", "Voxels")
        self._implicit_callback_ref: Any = None

    def _destroy_native(self) -> None:
        try:
            self.metadata.close()
        finally:
            Library._lib().Voxels_Destroy(self._handle)

    @classmethod
    def from_lattice(cls, lattice: Lattice) -> "Voxels":
        vox = cls()
        vox.render_lattice(lattice)
        return vox

    @classmethod
    def from_scalar_field(cls, field: "ScalarField") -> "Voxels":
        vox = cls()
        box_min, box_max = field.bounding_box()
        vox.render_implicit(box_min, box_max, field)
        return vox

    @classmethod
    def from_implicit(
        cls,
        implicit: IImplicit,
        bounds_min: Vector3Like,
        bounds_max: Vector3Like,
    ) -> "Voxels":
        vox = cls()
        vox.render_implicit(bounds_min, bounds_max, implicit)
        return vox

    @classmethod
    def from_bounded_implicit(cls, implicit: IBoundedImplicit) -> "Voxels":
        bounds_min, bounds_max = implicit.oBounds
        return cls.from_implicit(implicit, bounds_min, bounds_max)

    @classmethod
    def sphere(cls, center: Vector3Like, radius: float) -> "Voxels":
        with Lattice() as lattice:
            lattice.add_sphere(center, radius)
            return cls.from_lattice(lattice)

    voxSphere = sphere

    @classmethod
    def from_mesh(cls, mesh: Mesh) -> "Voxels":
        vox = cls()
        vox.render_mesh(mesh)
        return vox

    def duplicate(self) -> "Voxels":
        self._ensure_open()
        return Voxels(copy_from=self)

    voxDuplicate = duplicate

    def is_valid(self) -> bool:
        return bool(Library._lib().Voxels_bIsValid(self.handle))

    def bool_add(self, other: "Voxels") -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_BoolAdd(self.handle, other.handle)
        return self

    BoolAdd = bool_add

    def bool_add_all(self, items: Sequence["Voxels"]) -> "Voxels":
        for item in items:
            self.bool_add(item)
        return self

    BoolAddAll = bool_add_all

    def vox_bool_add(self, other: "Voxels") -> "Voxels":
        return self.duplicate().bool_add(other)

    voxBoolAdd = vox_bool_add

    def vox_bool_add_all(self, items: Sequence["Voxels"]) -> "Voxels":
        return self.duplicate().bool_add_all(items)

    voxBoolAddAll = vox_bool_add_all

    @staticmethod
    def voxCombine(vox1: "Voxels", vox2: "Voxels") -> "Voxels":
        return vox1.vox_bool_add(vox2)

    @staticmethod
    def voxCombineAll(items: Sequence["Voxels"]) -> "Voxels":
        vox = Voxels()
        return vox.bool_add_all(items)

    def bool_subtract(self, other: "Voxels") -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_BoolSubtract(self.handle, other.handle)
        return self

    BoolSubtract = bool_subtract

    def bool_subtract_all(self, items: Sequence["Voxels"]) -> "Voxels":
        for item in items:
            self.bool_subtract(item)
        return self

    BoolSubtractAll = bool_subtract_all

    def vox_bool_subtract(self, other: "Voxels") -> "Voxels":
        return self.duplicate().bool_subtract(other)

    voxBoolSubtract = vox_bool_subtract

    def vox_bool_subtract_all(self, items: Sequence["Voxels"]) -> "Voxels":
        return self.duplicate().bool_subtract_all(items)

    voxBoolSubtractAll = vox_bool_subtract_all

    def bool_intersect(self, other: "Voxels") -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_BoolIntersect(self.handle, other.handle)
        return self

    BoolIntersect = bool_intersect

    def vox_bool_intersect(self, other: "Voxels") -> "Voxels":
        return self.duplicate().bool_intersect(other)

    voxBoolIntersect = vox_bool_intersect

    def bool_add_smooth(self, other: "Voxels", smooth_distance: float) -> "Voxels":
        self._ensure_open()
        if not hasattr(Library._lib(), "Voxels_BoolAddSmooth"):
            raise NotImplementedError("Voxels_BoolAddSmooth is not available in this PicoGK runtime")
        Library._lib().Voxels_BoolAddSmooth(self.handle, other.handle, ctypes.c_float(smooth_distance))
        return self

    def offset(self, distance_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_Offset(self.handle, ctypes.c_float(distance_mm))
        return self

    Offset = offset

    def voxOffset(self, distance_mm: float) -> "Voxels":
        return self.duplicate().offset(distance_mm)

    def double_offset(self, first_mm: float, second_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_DoubleOffset(self.handle, ctypes.c_float(first_mm), ctypes.c_float(second_mm))
        return self

    DoubleOffset = double_offset

    def voxDoubleOffset(self, first_mm: float, second_mm: float) -> "Voxels":
        return self.duplicate().double_offset(first_mm, second_mm)

    def triple_offset(self, distance_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_TripleOffset(self.handle, ctypes.c_float(distance_mm))
        return self

    TripleOffset = triple_offset

    def voxTripleOffset(self, distance_mm: float) -> "Voxels":
        return self.duplicate().triple_offset(distance_mm)

    def Smoothen(self, distance_mm: float) -> "Voxels":
        return self.triple_offset(distance_mm)

    def voxSmoothen(self, distance_mm: float) -> "Voxels":
        return self.voxTripleOffset(distance_mm)

    def OverOffset(self, first_mm: float, final_surface_dist_mm: float = 0.0) -> "Voxels":
        return self.double_offset(first_mm, -(first_mm - final_surface_dist_mm))

    def voxOverOffset(self, first_mm: float, final_surface_dist_mm: float = 0.0) -> "Voxels":
        return self.duplicate().OverOffset(first_mm, final_surface_dist_mm)

    def Fillet(self, rounding_mm: float) -> "Voxels":
        return self.OverOffset(rounding_mm)

    def voxFillet(self, rounding_mm: float) -> "Voxels":
        return self.voxOverOffset(rounding_mm)

    def voxShell(self, neg_offset_mm: float, pos_offset_mm: float | None = None, smooth_inner_mm: float = 0.0) -> "Voxels":
        if pos_offset_mm is None:
            if neg_offset_mm < 0:
                return self.vox_bool_subtract(self.voxOffset(neg_offset_mm))
            return self.voxOffset(neg_offset_mm).vox_bool_subtract(self)
        if neg_offset_mm > pos_offset_mm:
            neg_offset_mm, pos_offset_mm = pos_offset_mm, neg_offset_mm
        inner = self.voxOffset(neg_offset_mm)
        if smooth_inner_mm > 0:
            inner = inner.voxTripleOffset(smooth_inner_mm)
        outer = self.voxOffset(pos_offset_mm)
        return outer.vox_bool_subtract(inner)

    def gaussian(self, size_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_Gaussian(self.handle, ctypes.c_float(size_mm))
        return self

    Gaussian = gaussian

    def median(self, size_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_Median(self.handle, ctypes.c_float(size_mm))
        return self

    Median = median

    def mean(self, size_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_Mean(self.handle, ctypes.c_float(size_mm))
        return self

    Mean = mean

    def render_mesh(self, mesh: Mesh) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_RenderMesh(self.handle, mesh.handle)
        return self

    RenderMesh = render_mesh

    def render_lattice(self, lattice: Lattice) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_RenderLattice(self.handle, lattice.handle)
        return self

    RenderLattice = render_lattice

    def render_implicit(
        self,
        bounds_min: Vector3Like,
        bounds_max: Vector3Like,
        implicit: IImplicit,
    ) -> "Voxels":
        self._ensure_open()
        bbox = as_bbox3(bounds_min, bounds_max)
        cb = _create_native_distance_callback(implicit)
        self._implicit_callback_ref = cb
        Library._lib().Voxels_RenderImplicit(self.handle, ctypes.byref(bbox), cb)
        return self

    RenderImplicit = render_implicit

    def render_bounded_implicit(self, implicit: IBoundedImplicit) -> "Voxels":
        bounds_min, bounds_max = implicit.oBounds
        return self.render_implicit(bounds_min, bounds_max, implicit)

    def intersect_implicit(self, implicit: IImplicit) -> "Voxels":
        self._ensure_open()
        cb = _create_native_distance_callback(implicit)
        self._implicit_callback_ref = cb
        Library._lib().Voxels_IntersectImplicit(self.handle, cb)
        return self

    IntersectImplicit = intersect_implicit

    def voxIntersectImplicit(self, implicit: IImplicit) -> "Voxels":
        return self.duplicate().intersect_implicit(implicit)

    def project_z_slice(self, start_mm: float, end_mm: float) -> "Voxels":
        self._ensure_open()
        Library._lib().Voxels_ProjectZSlice(self.handle, ctypes.c_float(start_mm), ctypes.c_float(end_mm))
        return self

    ProjectZSlice = project_z_slice

    def voxProjectZSlice(self, start_mm: float, end_mm: float) -> "Voxels":
        return self.duplicate().project_z_slice(start_mm, end_mm)

    def is_inside(self, point: Vector3Like) -> bool:
        self._ensure_open()
        p = as_vec3(point)
        lib = Library._lib()
        if hasattr(lib, "Voxels_bIsInside"):
            return bool(lib.Voxels_bIsInside(self.handle, ctypes.byref(p)))

        # Compatibility fallback for older runtimes that predate Voxels_bIsInside.
        # Sample signed distance from a scalar field built from this voxel set.
        from .fields import ScalarField
        with ScalarField.from_voxels(self) as field:
            return field.signed_distance(point) < 0.0

    bIsInside = is_inside

    def is_equal(self, other: "Voxels") -> bool:
        self._ensure_open()
        return bool(Library._lib().Voxels_bIsEqual(self.handle, other.handle))

    bIsEqual = is_equal

    def calculate_properties(self) -> tuple[float, tuple[tuple[float, float, float], tuple[float, float, float]]]:
        self._ensure_open()
        volume = ctypes.c_float(0.0)
        box = BBox3()
        Library._lib().Voxels_CalculateProperties(self.handle, ctypes.byref(volume), ctypes.byref(box))
        return float(volume.value), (vec3_tuple(box.min), vec3_tuple(box.max))

    CalculateProperties = calculate_properties

    def oCalculateBoundingBox(self) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        with Mesh.from_voxels(self) as mesh:
            return mesh.bounding_box()

    def get_surface_normal(self, point: Vector3Like) -> tuple[float, float, float]:
        self._ensure_open()
        p = as_vec3(point)
        out = Vec3()
        Library._lib().Voxels_GetSurfaceNormal(self.handle, ctypes.byref(p), ctypes.byref(out))
        return vec3_tuple(out)

    vecSurfaceNormal = get_surface_normal

    def closest_point_on_surface(self, search: Vector3Like) -> tuple[bool, tuple[float, float, float]]:
        self._ensure_open()
        s = as_vec3(search)
        out = Vec3()
        ok = bool(Library._lib().Voxels_bClosestPointOnSurface(self.handle, ctypes.byref(s), ctypes.byref(out)))
        return ok, vec3_tuple(out)

    bClosestPointOnSurface = closest_point_on_surface

    def vecClosestPointOnSurface(self, search: Vector3Like) -> tuple[float, float, float]:
        ok, point = self.closest_point_on_surface(search)
        if not ok:
            raise ValueError("Empty voxel field used in ClosestPointOnSurface")
        return point

    def ray_cast_to_surface(self, search: Vector3Like, direction: Vector3Like) -> tuple[bool, tuple[float, float, float]]:
        self._ensure_open()
        s = as_vec3(search)
        d = as_vec3(direction)
        out = Vec3()
        ok = bool(Library._lib().Voxels_bRayCastToSurface(self.handle, ctypes.byref(s), ctypes.byref(d), ctypes.byref(out)))
        return ok, vec3_tuple(out)

    bRayCastToSurface = ray_cast_to_surface

    def vecRayCastToSurface(self, search: Vector3Like, direction: Vector3Like) -> tuple[float, float, float]:
        ok, point = self.ray_cast_to_surface(search, direction)
        if not ok:
            raise ValueError("No intersection with surface in RayCastToSurface")
        return point

    def voxel_dimensions(self) -> VoxelDimensions:
        self._ensure_open()
        xo = ctypes.c_int()
        yo = ctypes.c_int()
        zo = ctypes.c_int()
        xs = ctypes.c_int()
        ys = ctypes.c_int()
        zs = ctypes.c_int()
        Library._lib().Voxels_GetVoxelDimensions(
            self.handle,
            ctypes.byref(xo),
            ctypes.byref(yo),
            ctypes.byref(zo),
            ctypes.byref(xs),
            ctypes.byref(ys),
            ctypes.byref(zs),
        )
        return VoxelDimensions(xo.value, yo.value, zo.value, xs.value, ys.value, zs.value)

    GetVoxelDimensions = voxel_dimensions

    def trim(self, oBox: tuple[Vector3Like, Vector3Like]) -> "Voxels":
        from .utils import Utils

        with Utils.mshCreateCube(oBox) as msh_trim:
            with Voxels.from_mesh(msh_trim) as vox_trim:
                self.bool_intersect(vox_trim)
        return self

    Trim = trim

    def vecZSliceOrigin(self, z_slice: int = 0) -> tuple[float, float, float]:
        dims = self.voxel_dimensions()
        return Library.voxels_to_mm((dims.x_origin, dims.y_origin, dims.z_origin + z_slice))

    def nSliceCount(self) -> int:
        return self.voxel_dimensions().z_size

    def slice(self, z_slice: int) -> tuple[np.ndarray, float]:
        self._ensure_open()
        dims = self.voxel_dimensions()
        count = dims.x_size * dims.y_size
        buf = (ctypes.c_float * count)()
        background = ctypes.c_float(0.0)
        Library._lib().Voxels_GetSlice(self.handle, int(z_slice), ctypes.cast(buf, ctypes.c_void_p), ctypes.byref(background))
        arr = np.frombuffer(buf, dtype=np.float32).reshape((dims.y_size, dims.x_size))
        return arr.copy(), float(background.value)

    def GetVoxelSlice(self, z_slice: int, mode: ESliceMode = ESliceMode.SignedDistance) -> np.ndarray:
        arr, background = self.slice(z_slice)
        if mode == ESliceMode.SignedDistance:
            return arr
        if mode == ESliceMode.BlackWhite:
            return np.where(arr <= 0, 0.0, 1.0).astype(np.float32)
        if mode == ESliceMode.Antialiased:
            if background <= 0:
                return np.where(arr <= 0, 0.0, 1.0).astype(np.float32)
            clipped = np.where(arr <= 0, 0.0, np.where(arr > background, 1.0, arr / background))
            return clipped.astype(np.float32)
        raise ValueError("mode must be 'signed_distance', 'black_white', or 'antialiased'")

    def interpolated_slice(self, z_slice: float) -> tuple[np.ndarray, float]:
        self._ensure_open()
        dims = self.voxel_dimensions()
        count = dims.x_size * dims.y_size
        buf = (ctypes.c_float * count)()
        background = ctypes.c_float(0.0)
        Library._lib().Voxels_GetInterpolatedSlice(self.handle, ctypes.c_float(z_slice), ctypes.cast(buf, ctypes.c_void_p), ctypes.byref(background))
        arr = np.frombuffer(buf, dtype=np.float32).reshape((dims.y_size, dims.x_size))
        return arr.copy(), float(background.value)

    def GetInterpolatedVoxelSlice(self, z_slice: float, mode: ESliceMode = ESliceMode.SignedDistance) -> np.ndarray:
        arr, background = self.interpolated_slice(z_slice)
        if mode == ESliceMode.SignedDistance:
            return arr
        if mode == ESliceMode.BlackWhite:
            return np.where(arr <= 0, 0.0, 1.0).astype(np.float32)
        if mode == ESliceMode.Antialiased:
            if background <= 0:
                return np.where(arr <= 0, 0.0, 1.0).astype(np.float32)
            clipped = np.where(arr <= 0, 0.0, np.where(arr > background, 1.0, arr / background))
            return clipped.astype(np.float32)
        raise ValueError("mode must be 'signed_distance', 'black_white', or 'antialiased'")

    def as_mesh(self) -> Mesh:
        self._ensure_open()
        return Mesh.from_voxels(self)

    mshAsMesh = as_mesh

    def oMetaData(self) -> FieldMetadata:
        return self.metadata

    @staticmethod
    def voxFromVdbFile(strFileName: str) -> "Voxels":
        from .openvdb import OpenVdbFile

        with OpenVdbFile(strFileName) as oFile:
            if oFile.nFieldCount() == 0:
                raise OSError(f"No fields contained in OpenVDB file {strFileName}")
            for n in range(oFile.nFieldCount()):
                if oFile.eFieldType(n) == OpenVdbFile.FieldType.VOXELS:
                    return oFile.voxGet(n)
        raise OSError(f"No voxel field (openvdb::GRID_LEVEL_SET) found in VDB file {strFileName}")

    def SaveToVdbFile(self, strFileName: str) -> None:
        from .openvdb import OpenVdbFile

        with OpenVdbFile() as oFile:
            oFile.nAdd(self)
            oFile.SaveToFile(strFileName)

    def voxVoxelizeHollow(self, fThickness: float) -> "Voxels":
        from .triangle_voxelization import TriangleVoxelization

        with self.mshAsMesh() as msh:
            return TriangleVoxelization.voxVoxelizeHollow(msh, fThickness)

    def oVectorize(self, fLayerHeight: float = 0.0, bUseAbsXYOrigin: bool = False) -> Any:
        from .image import ImageGrayScale
        from .slice import PolySlice, PolySliceStack

        voxel_size = float(getattr(Library, "voxel_size_mm", 0.5))
        if fLayerHeight == 0.0:
            fLayerHeight = voxel_size

        f_z_step = float(fLayerHeight) / voxel_size
        dims = self.GetVoxelDimensions()
        origin = (0.0, 0.0)
        if bUseAbsXYOrigin:
            origin = (dims.x_origin * voxel_size, dims.y_origin * voxel_size)

        slices: list[Any] = []
        started = False
        f_last_layer = float(dims.z_size - 1)
        f_z = 0.0
        f_layer_z = float(fLayerHeight)

        while f_z <= f_last_layer:
            arr = self.GetInterpolatedVoxelSlice(f_z, ESliceMode.SignedDistance)
            img = ImageGrayScale(dims.x_size, dims.y_size)
            img._data[:, :] = arr[:, :]

            poly_slice = PolySlice.oFromSdf(img, f_layer_z, origin, voxel_size)
            if not started and poly_slice.bIsEmpty():
                f_z += f_z_step
                f_layer_z += fLayerHeight
                continue

            started = True

            poly_slice.Close()
            slices.append(poly_slice)
            f_z += f_z_step
            f_layer_z += fLayerHeight

        if not slices:
            raise ValueError("Voxel field is empty - cannot write .CLI file")

        while slices and slices[-1].bIsEmpty():
            slices.pop()

        if not slices:
            raise ValueError("Voxel field is empty - cannot write .CLI file")

        return PolySliceStack(cast(Sequence[Any], slices))

    def SaveToCliFile(
        self,
        strFileName: str,
        fLayerHeight: float = 0.0,
        eFormat: Any = None,
        bUseAbsXYOrigin: bool = False,
    ) -> None:
        from .cli import Cli

        use_format = Cli.EFormat.FirstLayerWithContent if eFormat is None else eFormat
        oStack = self.oVectorize(fLayerHeight, bUseAbsXYOrigin)
        Cli.WriteSlicesToCliFile(cast(Any, oStack), strFileName, cast(Any, use_format))






from __future__ import annotations

import math
from collections.abc import Sequence
from enum import Enum

import numpy as np

from .._types import Vec3, Vector3Like, as_np3, as_vec3, clamp01, normalized
from ..splines import ISpline
from ..utils import SplineOperations
from ..utils import VecOperations
from .local_frames import LocalFrame


class Frames(ISpline):
    class EFrameType(Enum):
        CYLINDRICAL = "CYLINDRICAL"
        SPHERICAL = "SPHERICAL"
        Z = "Z"
        MIN_ROTATION = "MIN_ROTATION"

    def __init__(
        self,
        aPoints: list[np.ndarray],
        aLocalX: list[np.ndarray],
        aLocalY: list[np.ndarray],
        aLocalZ: list[np.ndarray],
    ) -> None:
        self.m_aPoints = aPoints
        self.m_aLocalX = aLocalX
        self.m_aLocalY = aLocalY
        self.m_aLocalZ = aLocalZ
        self.m_vecLastLocalX: np.ndarray | None = None

    @classmethod
    def from_length(
        cls,
        fLength: float,
        oConstLocalFrame: LocalFrame,
        fReparametrisationSpacing: float = 1.0,
    ) -> "Frames":
        start = as_np3(oConstLocalFrame.vecGetPosition())
        direction = as_np3(oConstLocalFrame.vecGetLocalZ())
        end = start + float(fLength) * direction

        points = [as_np3(p) for p in SplineOperations.aGetReparametrizedSpline([start, end], fReparametrisationSpacing)]
        local_z = [as_np3(oConstLocalFrame.vecGetLocalZ()) for _ in points]
        local_x = [as_np3(oConstLocalFrame.vecGetLocalX()) for _ in points]
        local_y = [as_np3(oConstLocalFrame.vecGetLocalY()) for _ in points]
        return cls(points, local_x, local_y, local_z)

    @classmethod
    def from_points_and_frame(
        cls,
        aPoints: Sequence[Vector3Like],
        oConstLocalFrame: LocalFrame,
        fReparametrisationSpacing: float = 1.0,
    ) -> "Frames":
        points = [as_np3(p) for p in SplineOperations.aGetReparametrizedSpline(aPoints, fReparametrisationSpacing)]
        local_z = [as_np3(oConstLocalFrame.vecGetLocalZ()) for _ in points]
        local_x = [as_np3(oConstLocalFrame.vecGetLocalX()) for _ in points]
        local_y = [as_np3(oConstLocalFrame.vecGetLocalY()) for _ in points]
        return cls(points, local_x, local_y, local_z)

    @classmethod
    def from_points_and_target_x(
        cls,
        aPoints: Sequence[Vector3Like],
        vecTargetX: Vector3Like,
        fReparametrisationSpacing: float = 1.0,
    ) -> "Frames":
        target_x = normalized(vecTargetX)
        points = [as_np3(p) for p in SplineOperations.aGetReparametrizedSpline(aPoints, fReparametrisationSpacing)]
        local_z = cls._tangent_directions(points)
        local_x = [cls.vecAlignWithTargetX(z, target_x) for z in local_z]
        local_y = [normalized(LocalFrame.vecBuildLocalY(z, x)) for z, x in zip(local_z, local_x)]

        n_samples = len(points)
        points = [as_np3(p) for p in SplineOperations.aGetNURBSpline(points, n_samples)]
        local_x = [as_np3(p) for p in SplineOperations.aGetNURBSpline(local_x, n_samples)]
        local_y = [as_np3(p) for p in SplineOperations.aGetNURBSpline(local_y, n_samples)]
        local_z = [as_np3(p) for p in SplineOperations.aGetNURBSpline(local_z, n_samples)]
        return cls(points, local_x, local_y, local_z)

    @classmethod
    def from_points_and_frame_type(
        cls,
        aPoints: Sequence[Vector3Like],
        eFrameType: "Frames.EFrameType",
        fReparametrisationSpacing: float = 1.0,
    ) -> "Frames":
        points = [as_np3(p) for p in SplineOperations.aGetReparametrizedSpline(aPoints, fReparametrisationSpacing)]
        local_z = cls._tangent_directions(points)

        local_x: list[np.ndarray] = []
        last_local_x: np.ndarray | None = None
        for point, z_vec in zip(points, local_z):
            x_vec, last_local_x = cls._align_with_frame_type(point, z_vec, eFrameType, last_local_x)
            local_x.append(x_vec)

        local_y = [normalized(LocalFrame.vecBuildLocalY(z, x)) for z, x in zip(local_z, local_x)]
        out = cls(points, local_x, local_y, local_z)
        out.m_vecLastLocalX = last_local_x
        return out

    @staticmethod
    def _align_with_frame_type(
        vecPt: np.ndarray,
        vecLocalZ: np.ndarray,
        eFrameType: "Frames.EFrameType",
        last_local_x: np.ndarray | None,
    ) -> tuple[np.ndarray, np.ndarray | None]:
        if eFrameType == Frames.EFrameType.MIN_ROTATION:
            if last_local_x is None:
                target_x = Frames.vecGetTargetX(vecPt, Frames.EFrameType.Z)
                last_local_x = Frames.vecAlignWithTargetX(vecLocalZ, target_x)
            local_x = Frames.vecAlignWithTargetX(vecLocalZ, last_local_x)
            return local_x, local_x

        target_x = Frames.vecGetTargetX(vecPt, eFrameType)
        local_x = Frames.vecAlignWithTargetX(vecLocalZ, target_x)
        return local_x, last_local_x

    @staticmethod
    def vecAlignWithTargetX(vecLocalZ: Vector3Like, vecTargetX: Vector3Like) -> np.ndarray:
        local_z = normalized(vecLocalZ)
        target_x = normalized(vecTargetX)
        init_x = VecOperations.vecGetOrthogonalDir(local_z)
        init_y = np.cross(init_x, local_z)

        max_dot = abs(float(np.dot(init_x, target_x)))
        final_x = init_x
        for angle in np.arange(0.0, 180.0, 0.01):
            phi = (2.0 * math.pi / 360.0) * float(angle)
            candidate = math.cos(phi) * init_x + math.sin(phi) * init_y
            dot_val = abs(float(np.dot(candidate, target_x)))
            if dot_val > max_dot:
                max_dot = dot_val
                final_x = candidate

        return normalized(VecOperations.vecFlipForAlignment(final_x, target_x))

    @staticmethod
    def vecGetTargetX(vecPt: Vector3Like, eFrameType: "Frames.EFrameType") -> np.ndarray:
        point = as_np3(vecPt)
        if eFrameType == Frames.EFrameType.CYLINDRICAL:
            return normalized((point[0], point[1], 0.0))
        if eFrameType == Frames.EFrameType.SPHERICAL:
            return normalized(point)
        return np.array([0.0, 0.0, 1.0], dtype=np.float64)

    @staticmethod
    def _tangent_directions(points: list[np.ndarray]) -> list[np.ndarray]:
        tangents: list[np.ndarray] = []
        for idx in range(1, len(points) - 1):
            tangents.append(normalized(points[idx] - points[idx - 1]))
        if not tangents:
            tangents = [np.array([0.0, 0.0, 1.0], dtype=np.float64)]
        tangents.insert(0, tangents[0])
        tangents.append(tangents[-1])
        return tangents

    def _interp(self, values: list[np.ndarray], fLengthRatio: float) -> np.ndarray:
        lr = clamp01(fLengthRatio)
        f_step = lr * max(0, len(values) - 1)
        i_lower = int(min(f_step, len(values) - 1))
        i_upper = int(min(f_step + 1.0, len(values) - 1))
        ds = f_step - i_lower
        return values[i_lower] + ds * (values[i_upper] - values[i_lower])

    def vecGetSpineAlongLength(self, fLengthRatio: float) -> Vec3:
        return as_vec3(self._interp(self.m_aPoints, fLengthRatio))

    def vec_get_spine_along_length(self, fLengthRatio: float) -> Vec3:
        return self.vecGetSpineAlongLength(fLengthRatio)

    def vecGetLocalXAlongLength(self, fLengthRatio: float) -> Vec3:
        return as_vec3(self._interp(self.m_aLocalX, fLengthRatio))

    def vec_get_local_x_along_length(self, fLengthRatio: float) -> Vec3:
        return self.vecGetLocalXAlongLength(fLengthRatio)

    def vecGetLocalYAlongLength(self, fLengthRatio: float) -> Vec3:
        return as_vec3(self._interp(self.m_aLocalY, fLengthRatio))

    def vec_get_local_y_along_length(self, fLengthRatio: float) -> Vec3:
        return self.vecGetLocalYAlongLength(fLengthRatio)

    def vecGetLocalZAlongLength(self, fLengthRatio: float) -> Vec3:
        return as_vec3(self._interp(self.m_aLocalZ, fLengthRatio))

    def vec_get_local_z_along_length(self, fLengthRatio: float) -> Vec3:
        return self.vecGetLocalZAlongLength(fLengthRatio)

    def oGetLocalFrame(self, fLengthRatio: float) -> LocalFrame:
        return LocalFrame(
            self.vecGetSpineAlongLength(fLengthRatio),
            self.vecGetLocalZAlongLength(fLengthRatio),
            self.vecGetLocalXAlongLength(fLengthRatio),
        )

    def aGetPoints(self, nSamples: int | None = 500) -> list[Vec3]:
        if nSamples is None:
            return [as_vec3(p) for p in self.m_aPoints]
        if nSamples <= 0:
            return []
        if nSamples == len(self.m_aPoints):
            return [as_vec3(p) for p in self.m_aPoints]
        return SplineOperations.aGetReparametrizedSpline(self.m_aPoints, int(nSamples))



from __future__ import annotations

import math
from enum import Enum

from picogk import Mesh

from .._types import Vec3, as_np3, as_vec3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from ..modulations import LineModulation
from .base_pipe import BasePipe


class BasePipeSegment(BasePipe):
	class EMethod(Enum):
		START_END = "START_END"
		MID_RANGE = "MID_RANGE"

	def __init__(
		self,
		aFrames: Frames,
		fInnerRadius: float,
		fOuterRadius: float,
		oStartOrMidModulation: LineModulation,
		oEndOrRangeModulation: LineModulation,
		eMethod: "BasePipeSegment.EMethod",
		*,
		_length_steps: int,
	) -> None:
		super().__init__(aFrames, fInnerRadius, fOuterRadius, _length_steps=_length_steps)
		self.m_eMethod = eMethod
		if self.m_eMethod == self.EMethod.START_END:
			self.m_oMidModulation = 0.5 * (oStartOrMidModulation + oEndOrRangeModulation)
			self.m_oRangeModulation = oEndOrRangeModulation - oStartOrMidModulation
		else:
			self.m_oMidModulation = oStartOrMidModulation
			self.m_oRangeModulation = oEndOrRangeModulation

	@classmethod
	def from_frame(
		cls,
		oFrame: LocalFrame,
		fLength: float,
		fInnerRadius: float,
		fOuterRadius: float,
		oStartOrMidModulation: LineModulation,
		oEndOrRangeModulation: LineModulation,
		eMethod: "BasePipeSegment.EMethod",
	) -> "BasePipeSegment":
		return cls(
			Frames.from_length(fLength, oFrame),
			fInnerRadius,
			fOuterRadius,
			oStartOrMidModulation,
			oEndOrRangeModulation,
			eMethod,
			_length_steps=5,
		)

	@classmethod
	def from_frames(
		cls,
		aFrames: Frames,
		fInnerRadius: float,
		fOuterRadius: float,
		oStartOrMidModulation: LineModulation,
		oEndOrRangeModulation: LineModulation,
		eMethod: "BasePipeSegment.EMethod",
	) -> "BasePipeSegment":
		return cls(
			aFrames,
			fInnerRadius,
			fOuterRadius,
			oStartOrMidModulation,
			oEndOrRangeModulation,
			eMethod,
			_length_steps=500,
		)

	def mshConstruct(self) -> Mesh:
		mesh = super().mshConstruct()
		self._add_start_surface(mesh)
		self._add_end_surface(mesh)
		return mesh

	def _add_start_surface(self, mesh: Mesh) -> None:
		pr = self._phi_ratio(0)
		for il in range(1, self.m_nLengthSteps):
			l1, l2 = self._length_ratio(il - 1), self._length_ratio(il)
			for ir in range(1, self.m_nRadialSteps):
				r1, r2 = self._radius_ratio(ir - 1), self._radius_ratio(ir)
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, pr, r1), self.vecGetSurfacePoint(l1, pr, r2), self.vecGetSurfacePoint(l2, pr, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, pr, r1), self.vecGetSurfacePoint(l2, pr, r2), self.vecGetSurfacePoint(l2, pr, r1))

	def _add_end_surface(self, mesh: Mesh) -> None:
		pr = self._phi_ratio(self.m_nPolarSteps - 1)
		for il in range(1, self.m_nLengthSteps):
			l1, l2 = self._length_ratio(il - 1), self._length_ratio(il)
			for ir in range(1, self.m_nRadialSteps):
				r1, r2 = self._radius_ratio(ir - 1), self._radius_ratio(ir)
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, pr, r1), self.vecGetSurfacePoint(l2, pr, r2), self.vecGetSurfacePoint(l1, pr, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, pr, r1), self.vecGetSurfacePoint(l2, pr, r1), self.vecGetSurfacePoint(l2, pr, r2))

	def vecGetSurfacePoint(self, fLengthRatio: float, fPhiRatio: float, fRadiusRatio: float) -> Vec3:
		spine = as_np3(self.m_aFrames.vecGetSpineAlongLength(fLengthRatio))
		lx = as_np3(self.m_aFrames.vecGetLocalXAlongLength(fLengthRatio))
		ly = as_np3(self.m_aFrames.vecGetLocalYAlongLength(fLengthRatio))
		phi = self.m_oMidModulation.fGetModulation(fLengthRatio) + (fPhiRatio - 0.5) * self.m_oRangeModulation.fGetModulation(fLengthRatio)
		outer = self.fGetOuterRadius(phi, fLengthRatio)
		inner = self.fGetInnerRadius(phi, fLengthRatio)
		radius = fRadiusRatio * (outer - inner) + inner
		pt = spine + (radius * math.cos(phi)) * lx + (radius * math.sin(phi)) * ly
		return self.m_fnTrafo(as_vec3(pt))


__all__ = ["BasePipeSegment"]



from __future__ import annotations

import numpy as np

from picogk import Voxels
from shape_kernel._types import Vector3Like, as_np3
from shape_kernel.utils import Uf

from .i_beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell


class CellBasedBeamThickness(IBeamThickness):
	def __init__(self, fMinBeamThickness: float, fMaxBeamThickness: float) -> None:
		self.m_fMinBeamThickness = float(fMinBeamThickness)
		self.m_fMaxBeamThickness = float(fMaxBeamThickness)
		self.m_xCell: IUnitCell | None = None

	def UpdateCell(self, xCell: IUnitCell) -> None:
		self.m_xCell = xCell

	def fGetBeamThickness(self, vecPt: Vector3Like) -> float:
		if self.m_xCell is None:
			raise ValueError("No Unit Cell specified.")

		vecCentre = as_np3(self.m_xCell.vecGetCellCentre())
		vecPoint = as_np3(vecPt)
		fDist = float(np.linalg.norm(vecPoint - vecCentre))
		vecSize = as_np3(self.m_xCell.oGetCellBounding().vecSize())
		fHalfDiag = 0.5 * float(np.linalg.norm(vecSize))
		if fHalfDiag <= 1e-12:
			fRatio = 0.0
		else:
			fRatio = min(max(fDist / fHalfDiag, 0.0), 1.0)
		return Uf.fTransFixed(self.m_fMinBeamThickness, self.m_fMaxBeamThickness, fRatio)

	def SetBoundingVoxels(self, voxBounding: Voxels) -> None:
		return None


__all__ = ["CellBasedBeamThickness"]


from __future__ import annotations

import numpy as np

from shape_kernel._types import Vec3, as_np3, as_vec3


class RobinsonTriangle:
    def __init__(self, vecA: Vec3, vecB: Vec3, vecC: Vec3) -> None:
        self.m_vecA: Vec3 = as_vec3(vecA)
        self.m_vecB: Vec3 = as_vec3(vecB)
        self.m_vecC: Vec3 = as_vec3(vecC)
        self.m_vecCentre: np.ndarray = 0.5 * (as_np3(self.m_vecA) + as_np3(self.m_vecC))

    def oGetFlippedTriangle(self) -> "RobinsonTriangle":
        vec_new_b = as_np3(self.m_vecB) + 2.0 * (self.m_vecCentre - as_np3(self.m_vecB))
        return RobinsonTriangle(self.m_vecA, as_vec3(vec_new_b), self.m_vecC)


__all__ = ["RobinsonTriangle"]



from __future__ import annotations

import math

from picogk import Mesh, Voxels

from .._types import Vec3, as_np3, as_vec3
from ..frames.local_frames import LocalFrame
from ..modulations import SurfaceModulation
from .base_shape import BaseShape


class BaseSphere(BaseShape):
	def __init__(self, oFrame: LocalFrame, fRadius: float = 10.0) -> None:
		super().__init__()
		self.SetAzimuthalSteps(360)
		self.SetPolarSteps(180)
		self.m_oFrame = oFrame
		self.m_oRadiusModulation = SurfaceModulation(fRadius)

	def SetRadius(self, oModulation: SurfaceModulation) -> None:
		self.m_oRadiusModulation = oModulation

	def SetAzimuthalSteps(self, nAzimuthalSteps: int) -> None:
		self.m_nAzimuthalSteps = int(nAzimuthalSteps)

	def SetPolarSteps(self, nPolarSteps: int) -> None:
		self.m_nPolarSteps = int(nPolarSteps)

	def voxConstruct(self) -> Voxels:
		return Voxels.from_mesh(self.mshConstruct())

	def mshConstruct(self) -> Mesh:
		mesh = Mesh()
		rr = 1.0
		for it in range(1, self.m_nAzimuthalSteps):
			t1 = (1.0 / (self.m_nAzimuthalSteps - 1)) * (it - 1)
			t2 = (1.0 / (self.m_nAzimuthalSteps - 1)) * it
			for ip in range(1, self.m_nPolarSteps):
				p1 = (1.0 / (self.m_nPolarSteps - 1)) * (ip - 1)
				p2 = (1.0 / (self.m_nPolarSteps - 1)) * ip
				mesh.nAddTriangle(self.vecGetSurfacePoint(t1, p1, rr), self.vecGetSurfacePoint(t2, p1, rr), self.vecGetSurfacePoint(t2, p2, rr))
				mesh.nAddTriangle(self.vecGetSurfacePoint(t1, p1, rr), self.vecGetSurfacePoint(t2, p2, rr), self.vecGetSurfacePoint(t1, p2, rr))
		return mesh

	def fGetRadius(self, fPhi: float, fLengthRatio: float) -> float:
		return self.m_oRadiusModulation.fGetModulation(fPhi, fLengthRatio)

	def vecGetSurfacePoint(self, fPhiRatio: float, fThetaRatio: float, fRadiusRatio: float) -> Vec3:
		theta = math.pi * fThetaRatio
		phi = 2.0 * math.pi * fPhiRatio
		fRadius = fRadiusRatio * self.fGetRadius(phi, theta)
		fx = fRadius * math.cos(phi) * math.sin(theta)
		fy = fRadius * math.sin(phi) * math.sin(theta)
		fz = fRadius * math.cos(theta)
		pos = as_np3(self.m_oFrame.vecGetPosition())
		lx = as_np3(self.m_oFrame.vecGetLocalX())
		ly = as_np3(self.m_oFrame.vecGetLocalY())
		lz = as_np3(self.m_oFrame.vecGetLocalZ())
		return self.m_fnTrafo(as_vec3(pos + fx * lx + fy * ly + fz * lz))


__all__ = ["BaseSphere"]



from .implicit_lidinoid import ImplicitLidinoid
from .implicit_modular import ImplicitModular
from .implicit_radial_gyroid import ImplicitRadialGyroid
from .implicit_randomized_schwarz_primitive import ImplicitRandomizedSchwarzPrimitive
from .implicit_schwarz_diamond import ImplicitSchwarzDiamond
from .implicit_schwarz_primitive import ImplicitSchwarzPrimitive
from .implicit_split_void_gyroid import ImplicitSplitVoidGyroid
from .implicit_split_wall_gyroid import ImplicitSplitWallGyroid

__all__ = [
	"ImplicitLidinoid",
	"ImplicitModular",
	"ImplicitRadialGyroid",
	"ImplicitRandomizedSchwarzPrimitive",
	"ImplicitSchwarzDiamond",
	"ImplicitSchwarzPrimitive",
	"ImplicitSplitVoidGyroid",
	"ImplicitSplitWallGyroid",
]


from .color_scale_2d import CustomColorScale2D, IColorScale, LinearColorScale2D, SmoothColorScale2D
from .color_scale_3d import ColorScale3D
from .mesh_painter import MeshPainter
from .rainbox_spectrum import ISpectrum, RainboxSpectrum
from .rotation_animator import RotationAnimator
from .color_palette import Cp

__all__ = [
    "Cp",
    "ColorScale3D",
    "CustomColorScale2D",
    "IColorScale",
    "ISpectrum",
    "LinearColorScale2D",
    "MeshPainter",
    "RainboxSpectrum",
    "RotationAnimator",
    "SmoothColorScale2D",
]



from __future__ import annotations

import time

from picogk import VedoViewer


class RotationAnimator:
	def __init__(self, oViewer: VedoViewer) -> None:
		self.m_oViewer = oViewer

	def Do(self, fStep: float = 0.5) -> None:
		fElevation = float(self.m_oViewer.m_fElevation)
		fCurrentOrbit = float(self.m_oViewer.m_fOrbit)
		self.m_oViewer.SetViewAngles(fCurrentOrbit + float(fStep), fElevation)

	def Run(self, nSteps: int = 360, fWaitSeconds: float = 0.01, fStep: float = 0.5) -> None:
		for _ in range(int(nSteps)):
			time.sleep(float(fWaitSeconds))
			self.Do(fStep)


__all__ = ["RotationAnimator"]



from __future__ import annotations

import math
from abc import ABC, abstractmethod

from shape_kernel._types import Vec3, as_np3, as_vec3

from .robinson_triangle import RobinsonTriangle


class ISubDTile(ABC):
    @abstractmethod
    def aGetInflatedSubTiles(self) -> list["ISubDTile"]:
        raise NotImplementedError

    @abstractmethod
    def aGetTileVertices(self) -> list[Vec3]:
        raise NotImplementedError


class RhombicTile(ISubDTile):
    m_fPsi = (math.sqrt(5.0) - 1.0) / 2.0

    def __init__(self, oTri: RobinsonTriangle) -> None:
        self.m_oTri_01 = oTri
        self.m_oTri_02 = oTri.oGetFlippedTriangle()

    @abstractmethod
    def aGetInflatedSubTiles(self) -> list[ISubDTile]:
        raise NotImplementedError

    def aGetTileVertices(self) -> list[Vec3]:
        return [
            self.m_oTri_01.m_vecA,
            self.m_oTri_01.m_vecB,
            self.m_oTri_01.m_vecC,
            self.m_oTri_02.m_vecB,
        ]


class SmallRhombicTile(RhombicTile):
    def aGetInflatedSubTiles(self) -> list[ISubDTile]:
        vec_d = as_np3(self.m_oTri_01.m_vecB) + self.m_fPsi * (as_np3(self.m_oTri_01.m_vecA) - as_np3(self.m_oTri_01.m_vecB))
        o_sub_tile_1: ISubDTile = SmallRhombicTile(RobinsonTriangle(as_vec3(vec_d), self.m_oTri_01.m_vecC, self.m_oTri_01.m_vecA))
        o_sub_tile_2: ISubDTile = LargeRhombicTile(RobinsonTriangle(self.m_oTri_01.m_vecC, as_vec3(vec_d), self.m_oTri_01.m_vecB))

        vec_d = as_np3(self.m_oTri_02.m_vecB) + self.m_fPsi * (as_np3(self.m_oTri_02.m_vecA) - as_np3(self.m_oTri_02.m_vecB))
        o_sub_tile_11: ISubDTile = SmallRhombicTile(RobinsonTriangle(as_vec3(vec_d), self.m_oTri_02.m_vecC, self.m_oTri_02.m_vecA))
        o_sub_tile_22: ISubDTile = LargeRhombicTile(RobinsonTriangle(self.m_oTri_02.m_vecC, as_vec3(vec_d), self.m_oTri_02.m_vecB))

        return [o_sub_tile_1, o_sub_tile_2, o_sub_tile_11, o_sub_tile_22]


class LargeRhombicTile(RhombicTile):
    def aGetInflatedSubTiles(self) -> list[ISubDTile]:
        vec_d = as_np3(self.m_oTri_01.m_vecA) + self.m_fPsi * (as_np3(self.m_oTri_01.m_vecB) - as_np3(self.m_oTri_01.m_vecA))
        vec_e = as_np3(self.m_oTri_01.m_vecA) + self.m_fPsi * (as_np3(self.m_oTri_01.m_vecC) - as_np3(self.m_oTri_01.m_vecA))
        o_sub_tile_1: ISubDTile = LargeRhombicTile(RobinsonTriangle(as_vec3(vec_e), as_vec3(vec_d), self.m_oTri_01.m_vecA))
        o_sub_tile_2: ISubDTile = SmallRhombicTile(RobinsonTriangle(as_vec3(vec_d), as_vec3(vec_e), self.m_oTri_01.m_vecB))
        o_sub_tile_3: ISubDTile = LargeRhombicTile(RobinsonTriangle(self.m_oTri_01.m_vecC, as_vec3(vec_e), self.m_oTri_01.m_vecB))

        vec_d = as_np3(self.m_oTri_02.m_vecA) + self.m_fPsi * (as_np3(self.m_oTri_02.m_vecB) - as_np3(self.m_oTri_02.m_vecA))
        vec_e = as_np3(self.m_oTri_02.m_vecA) + self.m_fPsi * (as_np3(self.m_oTri_02.m_vecC) - as_np3(self.m_oTri_02.m_vecA))
        o_sub_tile_11: ISubDTile = LargeRhombicTile(RobinsonTriangle(as_vec3(vec_e), as_vec3(vec_d), self.m_oTri_02.m_vecA))
        o_sub_tile_22: ISubDTile = SmallRhombicTile(RobinsonTriangle(as_vec3(vec_d), as_vec3(vec_e), self.m_oTri_02.m_vecB))
        o_sub_tile_33: ISubDTile = LargeRhombicTile(RobinsonTriangle(self.m_oTri_02.m_vecC, as_vec3(vec_e), self.m_oTri_02.m_vecB))

        return [o_sub_tile_1, o_sub_tile_2, o_sub_tile_3, o_sub_tile_11, o_sub_tile_22, o_sub_tile_33]


__all__ = ["ISubDTile", "RhombicTile", "SmallRhombicTile", "LargeRhombicTile"]



from __future__ import annotations

import math
import random
from enum import Enum
from typing import TYPE_CHECKING, cast

from .._types import Vec3, Vector3Like, as_np3, as_vec3

if TYPE_CHECKING:
	from ..splines import ControlPointSpline


class Uf:
	_random: random.Random = random.Random()
	_bspline: ControlPointSpline | None = None

	@classmethod
	def _get_bspline(cls) -> ControlPointSpline:
		if cls._bspline is None:
			from ..splines import ControlPointSpline

			ctrl_pts = [(0.0, 0.0, 0.0), (0.0, 0.0, 0.5), (1.0, 0.0, 0.5), (1.0, 0.0, 1.0)]
			cls._bspline = ControlPointSpline(ctrl_pts)
		return cls._bspline

	@staticmethod
	def Wait(fSeconds: float) -> None:
		import time

		time.sleep(float(fSeconds))

	@staticmethod
	def fTransFixed(fValue1: float, fValue2: float, fS: float) -> float:
		bspline = Uf._get_bspline()
		pt = bspline.vecGetPointAt(float(fS))
		ratio = float(pt[0])
		return float(fValue1) + ratio * (float(fValue2) - float(fValue1))

	@staticmethod
	def vecTransFixed(vecPt1: Vector3Like, vecPt2: Vector3Like, fS: float) -> Vec3:
		p1 = as_vec3(vecPt1)
		p2 = as_vec3(vecPt2)
		return (
			Uf.fTransFixed(p1[0], p2[0], fS),
			Uf.fTransFixed(p1[1], p2[1], fS),
			Uf.fTransFixed(p1[2], p2[2], fS),
		)

	@staticmethod
	def _fGetNormalizedTangensHyperbolicus(fS: float, fTransitionS: float, fSmooth: float) -> float:
		return float(0.5 + 0.5 * math.tanh((float(fS) - float(fTransitionS)) / float(fSmooth)))

	@staticmethod
	def fTransSmooth(fValue1: float, fValue2: float, fS: float, fTransitionS: float, fSmooth: float) -> float:
		t = Uf._fGetNormalizedTangensHyperbolicus(fS, fTransitionS, fSmooth)
		return float(fValue1) * (1.0 - t) + float(fValue2) * t

	@staticmethod
	def vecTransSmooth(vecPt1: Vector3Like, vecPt2: Vector3Like, fS: float, fTransitionS: float, fSmooth: float) -> Vec3:
		t = Uf._fGetNormalizedTangensHyperbolicus(fS, fTransitionS, fSmooth)
		p1 = as_np3(vecPt1)
		p2 = as_np3(vecPt2)
		return as_vec3(p1 * (1.0 - t) + p2 * t)

	@staticmethod
	def fLimitValue(fValue: float, fMin: float, fMax: float) -> float:
		return max(float(fMin), min(float(fMax), float(fValue)))

	@staticmethod
	def fGetRandomGaussian(fMean: float, fStdDev: float, oRandom: random.Random | None = None) -> float:
		rng = oRandom if oRandom is not None else Uf._random
		x1 = max(1.0 - rng.random(), 1e-300)
		x2 = 1.0 - rng.random()
		y1 = float(math.sqrt(-2.0 * math.log(x1)) * math.cos(2.0 * math.pi * x2))
		return y1 * float(fStdDev) + float(fMean)

	@staticmethod
	def fGetRandomLinear(fMin: float, fMax: float, oRandom: random.Random | None = None) -> float:
		rng = oRandom if oRandom is not None else Uf._random
		return float(fMin) + (float(fMax) - float(fMin)) * rng.random()

	@staticmethod
	def bGetRandomBool(oRandom: random.Random | None = None) -> bool:
		return Uf.fGetRandomLinear(0.0, 1.0, oRandom) > 0.5

	@staticmethod
	def aGetFibonacciCirlePoints(fOuterRadius: float, nSamples: int) -> list[Vec3]:
		pts: list[Vec3] = []
		r_out = float(fOuterRadius)
		golden = math.pi * (1.0 + math.sqrt(5.0))
		for i in range(int(nSamples)):
			k = i + 0.5
			r = math.sqrt(k / float(nSamples))
			phi = golden * k
			pts.append((r_out * r * math.cos(phi), r_out * r * math.sin(phi), 0.0))
		return pts

	@staticmethod
	def aGetFibonacciSpherePoints(fOuterRadius: float, nSamples: int) -> list[Vec3]:
		pts: list[Vec3] = []
		r_out = float(fOuterRadius)
		golden = math.pi * (1.0 + math.sqrt(5.0))
		for i in range(int(nSamples)):
			k = i + 0.5
			phi_a = math.acos(1.0 - 2.0 * k / float(nSamples))
			theta = golden * k
			pts.append(
				(
					r_out * math.cos(theta) * math.sin(phi_a),
					r_out * math.sin(theta) * math.sin(phi_a),
					r_out * math.cos(phi_a),
				)
			)
		return pts

	class ESuperShape(Enum):
		ROUND = 0
		HEX = 1
		QUAD = 2
		TRI = 3

	@staticmethod
	def fGetSuperShapeRadius(fPhi: float, fM_or_shape: float | Uf.ESuperShape, fN1: float = 1.0, fN2: float = 1.0, fN3: float = 1.0) -> float:
		if isinstance(fM_or_shape, Uf.ESuperShape):
			e = fM_or_shape
			if e == Uf.ESuperShape.HEX:
				return Uf.fGetSuperShapeRadius(fPhi, 6.0, 2.0, 1.2, 1.2)
			if e == Uf.ESuperShape.QUAD:
				return Uf.fGetSuperShapeRadius(fPhi, 4.0, 20.0, 15.0, 15.0)
			if e == Uf.ESuperShape.TRI:
				return Uf.fGetSuperShapeRadius(fPhi, 3.0, 3.0, 4.0, 4.0)
			return 1.0
		fM = float(cast(float, fM_or_shape))
		a = math.pow(abs(math.cos(0.25 * fM * float(fPhi))), float(fN2))
		b = math.pow(abs(math.sin(0.25 * fM * float(fPhi))), float(fN3))
		return float(math.pow(a + b, -1.0 / float(fN1)))

	class EPolygon(Enum):
		HEX = 0
		QUAD = 1
		TRI = 2

	@staticmethod
	def fGetPolygonRadius(fPhi: float, nM_or_polygon: int | Uf.EPolygon) -> float:
		if isinstance(nM_or_polygon, Uf.EPolygon):
			e = nM_or_polygon
			if e == Uf.EPolygon.HEX:
				return Uf.fGetPolygonRadius(fPhi, 6)
			if e == Uf.EPolygon.QUAD:
				return Uf.fGetPolygonRadius(fPhi, 4)
			if e == Uf.EPolygon.TRI:
				return Uf.fGetPolygonRadius(fPhi, 3)
			return 1.0
		nM = int(cast(int, nM_or_polygon))
		d_phi = float(fPhi) % (2.0 * math.pi / float(nM))
		return float(math.cos(math.pi / float(nM)) / math.cos(d_phi - math.pi / float(nM)))


__all__ = ["Uf"]



from __future__ import annotations

import numpy as np
from .control_point_spline import ControlPointSpline

from .._types import Vec3, Vector3Like, as_np3, as_vec3
from .i_spline import ISpline


class CylindricalControlSpline(ISpline):
    class EDirection:
        RADIAL = "RADIAL"
        TANGENTIAL = "TANGENTIAL"
        Z = "Z"

    def __init__(self, vecStart: Vector3Like) -> None:
        self.m_aControlPoints: list[np.ndarray] = [as_np3(vecStart)]

    def AddRelativeStep(self, eDir: str, fStepLength: float) -> None:
        last_pos = self.m_aControlPoints[-1]
        if eDir == self.EDirection.Z:
            new_pos = last_pos + float(fStepLength) * np.array([0.0, 0.0, 1.0], dtype=np.float64)
            self.m_aControlPoints.append(new_pos)
            return
        radial_dir = np.array([last_pos[0], last_pos[1], 0.0], dtype=np.float64)
        radial_norm = float(np.linalg.norm(radial_dir))
        if radial_norm <= 1e-12:
            radial_dir = np.array([1.0, 0.0, 0.0], dtype=np.float64)
        else:
            radial_dir = radial_dir / radial_norm

        if eDir == self.EDirection.RADIAL:
            new_pos = last_pos + float(fStepLength) * radial_dir
        else:
            tangential = np.cross(np.array([0.0, 0.0, 1.0], dtype=np.float64), radial_dir)
            tangential_norm = float(np.linalg.norm(tangential))
            if tangential_norm <= 1e-12:
                tangential = np.array([0.0, 1.0, 0.0], dtype=np.float64)
            else:
                tangential = tangential / tangential_norm
            new_pos = last_pos + float(fStepLength) * tangential
        self.m_aControlPoints.append(new_pos)

    def AddAbsoluteStep(self, eDir: str, fNewValue: float) -> None:
        last_pos = self.m_aControlPoints[-1].copy()
        if eDir == self.EDirection.Z:
            last_pos[2] = float(fNewValue)
            self.m_aControlPoints.append(last_pos)
            return
        if eDir == self.EDirection.RADIAL:
            radial = np.array([last_pos[0], last_pos[1]], dtype=np.float64)
            norm = float(np.linalg.norm(radial))
            if norm <= 1e-12:
                last_pos[0] = float(fNewValue)
                last_pos[1] = 0.0
            else:
                scale = float(fNewValue) / norm
                last_pos[0] *= scale
                last_pos[1] *= scale
            self.m_aControlPoints.append(last_pos)

    def aGetPoints(self, nSamples: int | None = 500) -> list[Vec3]:
        if nSamples is None:
            nSamples = 500
        bspline = ControlPointSpline([as_vec3(p) for p in self.m_aControlPoints])
        return bspline.aGetPoints(nSamples)




from __future__ import annotations


import threading
from typing import Callable

from .._core import Library
from .vedo_viewer import VedoViewer

def go(
    voxel_size_mm: float,
    task: Callable[[], None],
    *,
    end_on_task_completion: bool = False,
    viewer: VedoViewer | None = None,
) -> None:
    """Initialize runtime, execute task in worker thread, then clean up."""

    errors: list[BaseException] = []

    def _runner() -> None:
        try:
            task()
        except BaseException as exc:  # noqa: BLE001
            errors.append(exc)

    Library.init(float(voxel_size_mm))
    try:
        Library.SetViewer(viewer)
        # For interactive viewer mode we render on the main thread after task completion.
        # Starting the background loop in that mode can race window lifetime on some backends.
        if viewer is not None and end_on_task_completion:
            viewer.start()

        worker = threading.Thread(target=_runner, daemon=True)
        worker.start()

        worker.join()

        if errors:
            raise errors[0]

        if viewer is not None and not end_on_task_completion:
            viewer.interact()
    finally:
        if viewer is not None:
            viewer.close()
        Library.ClearViewer()
        Library.destroy()




from __future__ import annotations

import math
from collections.abc import Sequence

import numpy as np

from .._types import Vec3, Vector3Like, as_np3, as_vec3
from .vec_operations import VecOperations


class LineDecimation:
	def __init__(self, aPoints: Sequence[Vector3Like], fMaxError: float) -> None:
		self._pts = [as_np3(p) for p in aPoints]
		i_start = 0
		indices: list[int] = [0]
		while i_start < len(self._pts) - 1:
			i_end = self._iGetNextIndex(i_start, float(fMaxError))
			i_start = i_end
			indices.append(i_start)
		self._decimated: list[Vec3] = [as_vec3(self._pts[i]) for i in indices]

	def _iGetNextIndex(self, i_start: int, fMaxError: float) -> int:
		if i_start > len(self._pts) - 2:
			return len(self._pts) - 1
		i_end = i_start + 1
		line_pts = list(self._pts[i_start : i_start + 2])
		for i_idx in range(i_start + 2, len(self._pts)):
			line_pts.append(self._pts[i_idx])
			if LineDecimation._fGetMaxError(line_pts) > fMaxError:
				break
			i_end = i_idx
		return i_end

	@staticmethod
	def _fGetMaxError(line_pts: list[np.ndarray]) -> float:
		line_dir = line_pts[-1] - line_pts[0]
		max_err = 0.0
		for i in range(1, len(line_pts) - 1):
			pointer = line_pts[i] - line_pts[0]
			angle = VecOperations.fGetAngleBetween(pointer, line_dir)
			err = math.sin(angle) * float(np.linalg.norm(pointer))
			if err > max_err:
				max_err = err
		return max_err

	def aGetDecimatedPoints(self) -> list[Vec3]:
		return list(self._decimated)


__all__ = ["LineDecimation"]



from .coordinate_trafo import CombinedTrafo, FunctionalScaleTrafo, ICoordinateTrafo, RadialTrafo, ScaleTrafo
from picogk import IImplicit
from .implicit_modular import ImplicitModular
from .random_deformation_field import RandomDeformationField
from .raw_tpms_patterns import (
	IRawTPMSPattern,
	RawGyroidTPMSPattern,
	RawLidinoidTPMSPattern,
	RawSchwarzDiamondTPMSPattern,
	RawSchwarzPrimitiveTPMSPattern,
	RawTransitionTPMSPattern,
)
from .splitting_logic import (
	FullVoidLogic,
	FullWallLogic,
	ISplittingLogic,
	NegativeHalfWallLogic,
	NegativeVoidLogic,
	PositiveHalfWallLogic,
	PositiveVoidLogic,
)
from .TPMSPresets import (
	ImplicitLidinoid,
	ImplicitRadialGyroid,
	ImplicitRandomizedSchwarzPrimitive,
	ImplicitSchwarzDiamond,
	ImplicitSchwarzPrimitive,
	ImplicitSplitVoidGyroid,
	ImplicitSplitWallGyroid,
)

__all__ = [
	"IImplicit",
	"ICoordinateTrafo",
	"ScaleTrafo",
	"FunctionalScaleTrafo",
	"RadialTrafo",
	"CombinedTrafo",
	"ISplittingLogic",
	"FullWallLogic",
	"FullVoidLogic",
	"PositiveHalfWallLogic",
	"NegativeHalfWallLogic",
	"PositiveVoidLogic",
	"NegativeVoidLogic",
	"IRawTPMSPattern",
	"RawGyroidTPMSPattern",
	"RawLidinoidTPMSPattern",
	"RawSchwarzPrimitiveTPMSPattern",
	"RawSchwarzDiamondTPMSPattern",
	"RawTransitionTPMSPattern",
	"RandomDeformationField",
	"ImplicitModular",
	"ImplicitLidinoid",
	"ImplicitRadialGyroid",
	"ImplicitRandomizedSchwarzPrimitive",
	"ImplicitSchwarzDiamond",
	"ImplicitSchwarzPrimitive",
	"ImplicitSplitVoidGyroid",
	"ImplicitSplitWallGyroid",
]


from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from picogk import BBox3, Lattice
from shape_kernel._types import Vec3, Vector3Like, as_np3, as_vec3

if TYPE_CHECKING:
	from .beam_thickness.i_beam_thickness import IBeamThickness


class _BBoxMinMax(Protocol):
	x: float
	y: float
	z: float


class _BBoxWithVecMinMax(Protocol):
	vecMin: Vector3Like
	vecMax: Vector3Like


class _BBoxWithMinMax(Protocol):
	min: _BBoxMinMax
	max: _BBoxMinMax


BBox3Like = BBox3 | tuple[Vector3Like, Vector3Like] | _BBoxWithVecMinMax | _BBoxWithMinMax


def _coerce_bbox3(box: BBox3Like) -> BBox3:
	if isinstance(box, BBox3):
		return box
	if isinstance(box, tuple) and len(box) == 2:
		return BBox3(as_vec3(box[0]), as_vec3(box[1]))
	if hasattr(box, "vecMin") and hasattr(box, "vecMax"):
		return BBox3(as_vec3(getattr(box, "vecMin")), as_vec3(getattr(box, "vecMax")))
	if hasattr(box, "min") and hasattr(box, "max"):
		return BBox3(
			(float(box.min.x), float(box.min.y), float(box.min.z)), # type: ignore
			(float(box.max.x), float(box.max.y), float(box.max.z)), # type: ignore
		)
	raise TypeError("Unsupported bounding box representation")


def _interpolate_point(vecPt1: Vector3Like, vecPt2: Vector3Like, ratio: float) -> Vec3:
	pt1 = as_np3(vecPt1)
	pt2 = as_np3(vecPt2)
	return as_vec3(pt1 + float(ratio) * (pt2 - pt1))


def _add_sampled_beam(oLattice: Lattice, vecPt1: Vector3Like, vecPt2: Vector3Like, xBeamThickness: IBeamThickness, nSamples: int = 2) -> None:
	n_samples = max(2, int(nSamples))
	if n_samples == 2:
		fRadius1 = 0.5 * float(xBeamThickness.fGetBeamThickness(vecPt1))
		fRadius2 = 0.5 * float(xBeamThickness.fGetBeamThickness(vecPt2))
		oLattice.AddBeam(vecPt1, vecPt2, fRadius1, fRadius2)
		return

	for i in range(1, n_samples):
		vecPt11 = _interpolate_point(vecPt1, vecPt2, float(i - 1) / float(n_samples - 1))
		vecPt22 = _interpolate_point(vecPt1, vecPt2, float(i) / float(n_samples - 1))
		fRadius11 = 0.5 * float(xBeamThickness.fGetBeamThickness(vecPt11))
		fRadius22 = 0.5 * float(xBeamThickness.fGetBeamThickness(vecPt22))
		oLattice.AddBeam(vecPt11, vecPt22, fRadius11, fRadius22)


def _add_polyline_beam(oLattice: Lattice, aPoints: list[Vec3], xBeamThickness: IBeamThickness) -> None:
	for i in range(1, len(aPoints)):
		vecPt11 = aPoints[i - 1]
		vecPt22 = aPoints[i]
		fRadius11 = 0.5 * float(xBeamThickness.fGetBeamThickness(vecPt11))
		fRadius22 = 0.5 * float(xBeamThickness.fGetBeamThickness(vecPt22))
		oLattice.AddBeam(vecPt11, vecPt22, fRadius11, fRadius22)


__all__ = ["_add_polyline_beam", "_add_sampled_beam", "_coerce_bbox3"]


from __future__ import annotations

import numpy as np

from picogk import BBox3
from shape_kernel import Cp, Sh
from shape_kernel._types import Vec3, Vector3Like, as_np3, as_vec3
from .i_unit_cell import IUnitCell


class CuboidCell(IUnitCell):
	def __init__(
		self,
		vecCorner_01: Vector3Like,
		vecCorner_02: Vector3Like,
		vecCorner_03: Vector3Like,
		vecCorner_04: Vector3Like,
		vecCorner_05: Vector3Like,
		vecCorner_06: Vector3Like,
		vecCorner_07: Vector3Like,
		vecCorner_08: Vector3Like,
	) -> None:
		self.m_aCornerPoints: list[Vec3] = [
			as_vec3(vecCorner_01),
			as_vec3(vecCorner_02),
			as_vec3(vecCorner_03),
			as_vec3(vecCorner_04),
			as_vec3(vecCorner_05),
			as_vec3(vecCorner_06),
			as_vec3(vecCorner_07),
			as_vec3(vecCorner_08),
		]

		corners = np.stack([as_np3(vecCorner) for vecCorner in self.m_aCornerPoints], axis=0)
		self.m_vecCentre = as_vec3(np.mean(corners, axis=0))
		self.m_oBBox = BBox3(as_vec3(np.min(corners, axis=0)), as_vec3(np.max(corners, axis=0)))

	def vecGetCellCentre(self) -> Vec3:
		return self.m_vecCentre

	def oGetCellBounding(self) -> BBox3:
		return self.m_oBBox

	def aGetCornerPoints(self) -> list[Vec3]:
		return self.m_aCornerPoints

	def PreviewUnitCell(self) -> None:
		Sh.PreviewLine([self.m_aCornerPoints[0], self.m_aCornerPoints[1], self.m_aCornerPoints[2], self.m_aCornerPoints[3], self.m_aCornerPoints[0]], Cp.clrBlack)
		Sh.PreviewLine([self.m_aCornerPoints[4], self.m_aCornerPoints[5], self.m_aCornerPoints[6], self.m_aCornerPoints[7], self.m_aCornerPoints[4]], Cp.clrBlack)
		Sh.PreviewLine([self.m_aCornerPoints[0], self.m_aCornerPoints[4]], Cp.clrBlack)
		Sh.PreviewLine([self.m_aCornerPoints[1], self.m_aCornerPoints[5]], Cp.clrBlack)
		Sh.PreviewLine([self.m_aCornerPoints[2], self.m_aCornerPoints[6]], Cp.clrBlack)
		Sh.PreviewLine([self.m_aCornerPoints[3], self.m_aCornerPoints[7]], Cp.clrBlack)


__all__ = ["CuboidCell"]


from __future__ import annotations

import math
from abc import ABC, abstractmethod

from shape_kernel.utils import Uf


class IRawTPMSPattern(ABC):
	@abstractmethod
	def fGetSignedDistance(self, fX: float, fY: float, fZ: float) -> float:
		raise NotImplementedError


class RawGyroidTPMSPattern(IRawTPMSPattern):
	m_fFrequencyScale = 2.0 * math.pi

	def fGetSignedDistance(self, fX: float, fY: float, fZ: float) -> float:
		return (
			math.sin(self.m_fFrequencyScale * fX) * math.cos(self.m_fFrequencyScale * fY)
			+ math.sin(self.m_fFrequencyScale * fY) * math.cos(self.m_fFrequencyScale * fZ)
			+ math.sin(self.m_fFrequencyScale * fZ) * math.cos(self.m_fFrequencyScale * fX)
		)


class RawLidinoidTPMSPattern(IRawTPMSPattern):
	m_fFrequencyScale = 0.5 * (2.0 * math.pi)

	def fGetSignedDistance(self, fX: float, fY: float, fZ: float) -> float:
		return (
			+0.5
			* (
				math.sin(2.0 * self.m_fFrequencyScale * fX) * math.cos(self.m_fFrequencyScale * fY) * math.sin(self.m_fFrequencyScale * fZ)
				+ math.sin(2.0 * self.m_fFrequencyScale * fY) * math.cos(self.m_fFrequencyScale * fZ) * math.sin(self.m_fFrequencyScale * fX)
				+ math.sin(2.0 * self.m_fFrequencyScale * fZ) * math.cos(self.m_fFrequencyScale * fX) * math.sin(self.m_fFrequencyScale * fY)
			)
			- 0.5
			* (
				math.cos(2.0 * self.m_fFrequencyScale * fX) * math.cos(2.0 * self.m_fFrequencyScale * fY)
				+ math.cos(2.0 * self.m_fFrequencyScale * fY) * math.cos(2.0 * self.m_fFrequencyScale * fZ)
				+ math.cos(2.0 * self.m_fFrequencyScale * fZ) * math.cos(2.0 * self.m_fFrequencyScale * fX)
			)
		)


class RawSchwarzPrimitiveTPMSPattern(IRawTPMSPattern):
	m_fFrequencyScale = 2.0 * math.pi

	def fGetSignedDistance(self, fX: float, fY: float, fZ: float) -> float:
		return (
			math.cos(self.m_fFrequencyScale * fX)
			+ math.cos(self.m_fFrequencyScale * fY)
			+ math.cos(self.m_fFrequencyScale * fZ)
		)


class RawSchwarzDiamondTPMSPattern(IRawTPMSPattern):
	m_fFrequencyScale = 0.5 * (2.0 * math.pi)

	def fGetSignedDistance(self, fX: float, fY: float, fZ: float) -> float:
		return (
			math.cos(self.m_fFrequencyScale * fX) * math.cos(self.m_fFrequencyScale * fY) * math.cos(self.m_fFrequencyScale * fZ)
			- math.sin(self.m_fFrequencyScale * fX) * math.sin(self.m_fFrequencyScale * fY) * math.sin(self.m_fFrequencyScale * fZ)
		)


class RawTransitionTPMSPattern(IRawTPMSPattern):
	def __init__(self) -> None:
		self.m_xTPMS_01 = RawSchwarzDiamondTPMSPattern()
		self.m_xTPMS_02 = RawSchwarzPrimitiveTPMSPattern()

	def fGetSignedDistance(self, fX: float, fY: float, fZ: float) -> float:
		fDist_01 = self.m_xTPMS_01.fGetSignedDistance(fX, fY, fZ)
		fDist_02 = self.m_xTPMS_02.fGetSignedDistance(fX, fY, fZ)
		fRatio = min(max((fX + 2.0) / 5.0, 0.0), 1.0)
		return Uf.fTransFixed(fDist_01, fDist_02, fRatio)


__all__ = [
	"IRawTPMSPattern",
	"RawGyroidTPMSPattern",
	"RawLidinoidTPMSPattern",
	"RawSchwarzPrimitiveTPMSPattern",
	"RawSchwarzDiamondTPMSPattern",
	"RawTransitionTPMSPattern",
]



import math
from pathlib import Path

import numpy as np

from _common.types import ColorLike, Vector3Like
from .polyline import PolyLine
from .image_io import TgaIo
from .._viewer_protocol import IViewer
from .._common import to_rgba
from picogk._core.voxels import Voxels
from .image import ImageColor
from .fields import ScalarField, VectorField


class SdfVisualizer:
    @staticmethod
    def imgEncodeFromSdf(
        oField: ScalarField,
        fBackgroundValue: float,
        nSlice: int,
        _clrBackground: ColorLike | None = None,
        _clrSurface: ColorLike | None = None,
        _clrInside: ColorLike | None = None,
        _clrOutside: ColorLike | None = None,
        _clrDefect: ColorLike | None = None,
    ) -> ImageColor:
        dims = oField.GetVoxelDimensions()
        img = ImageColor(dims.x_size, dims.y_size)
        if nSlice >= dims.z_size:
            return img
        colors = {
            "bg": to_rgba(_clrBackground or (0.0, 0.4, 1.0, 1.0)),
            "surf": to_rgba(_clrSurface or (1.0, 1.0, 1.0, 1.0)),
            "in": to_rgba(_clrInside or (0.8, 0.2, 1.0, 1.0)),
            "out": to_rgba(_clrOutside or (0.2, 0.8, 0.2, 1.0)),
            "def": to_rgba(_clrDefect or (1.0, 0.33, 0.0, 1.0)),
        }
        sdf = oField.GetVoxelSlice(nSlice)
        for y in range(dims.y_size):
            for x in range(dims.x_size):
                f = float(sdf[y, x])
                if not math.isfinite(f):
                    clr = colors["def"]
                elif abs(f) < np.finfo(np.float32).eps:
                    clr = colors["surf"]
                elif abs(f - fBackgroundValue) < np.finfo(np.float32).eps:
                    clr = colors["bg"]
                elif f < 0.0:
                    t = min(1.0, abs(f) / max(abs(fBackgroundValue), 1e-6))
                    base = colors["in"]
                    clr = (base[0] * (1.0 - 0.35 * t), base[1] * (1.0 - 0.35 * t), base[2] * (1.0 - 0.35 * t), base[3])
                else:
                    t = min(1.0, abs(f) / max(abs(fBackgroundValue), 1e-6))
                    base = colors["out"]
                    clr = (base[0] * (1.0 - 0.35 * t), base[1] * (1.0 - 0.35 * t), base[2] * (1.0 - 0.35 * t), base[3])
                img.SetValue(x, y, clr)
        return img

    @staticmethod
    def bDoesSliceContainDefect(oField: ScalarField, nSlice: int) -> bool:
        dims = oField.GetVoxelDimensions()
        if nSlice >= dims.z_size:
            return False
        sdf = oField.GetVoxelSlice(nSlice)
        return bool(np.any(~np.isfinite(sdf)))

    @staticmethod
    def bVisualizeSdfSlicesAsTgaStack(
        oField: ScalarField,
        fBackgroundValue: float,
        strPath: str,
        strFilePrefix: str = "Sdf_",
        bOnlyDefective: bool = False,
        _clrBackground: ColorLike | None = None,
        _clrSurface: ColorLike | None = None,
        _clrInside: ColorLike | None = None,
        _clrOutside: ColorLike | None = None,
        _clrDefect: ColorLike | None = None,
    ) -> bool:
        path = Path(strPath)
        path.mkdir(parents=True, exist_ok=True)
        dims = oField.GetVoxelDimensions()
        found_defect = False
        for s in range(dims.z_size):
            has_defect = SdfVisualizer.bDoesSliceContainDefect(oField, s)
            found_defect = found_defect or has_defect
            if bOnlyDefective and not has_defect:
                continue
            img = SdfVisualizer.imgEncodeFromSdf(
                oField,
                fBackgroundValue,
                s,
                _clrBackground,
                _clrSurface,
                _clrInside,
                _clrOutside,
                _clrDefect,
            )
            TgaIo.SaveTga(path / f"{strFilePrefix}{s:05d}.tga", img)
        return found_defect


class ActiveVoxelCounterScalar:
    @staticmethod
    def nCount(oField: ScalarField) -> int:
        count = 0

        def _visit(_: tuple[float, float, float], __: float) -> None:
            nonlocal count
            count += 1

        oField.TraverseActive(_visit)
        return count


class SurfaceNormalFieldExtractor:
    @staticmethod
    def oExtract(
        vox: Voxels,
        fSurfaceThresholdVx: float = 0.5,
        vecDirectionFilter: Vector3Like | None = None,
        fDirectionFilterTolerance: float = 0.0,
        vecScaleBy: Vector3Like | None = None,
    ) -> VectorField:
        direction = np.array(vecDirectionFilter if vecDirectionFilter is not None else (0.0, 0.0, 0.0), dtype=np.float64)
        if np.linalg.norm(direction) > 1e-12:
            direction = direction / np.linalg.norm(direction)
        scale_by = np.array(vecScaleBy if vecScaleBy is not None else (1.0, 1.0, 1.0), dtype=np.float64)

        src = ScalarField.from_voxels(vox)
        dst = VectorField()

        def _visit(pos: tuple[float, float, float], value: float) -> None:
            if abs(value) > fSurfaceThresholdVx:
                return
            normal = np.array(vox.vecSurfaceNormal(pos), dtype=np.float64)
            nlen = np.linalg.norm(normal)
            if nlen <= 1e-12:
                return
            normal = normal / nlen
            if np.linalg.norm(direction) > 1e-12:
                dev = abs(1.0 - float(np.dot(normal, direction)))
                if dev > float(fDirectionFilterTolerance):
                    return
            v = normal * scale_by
            dst.SetValue(pos, (float(v[0]), float(v[1]), float(v[2])))

        src.TraverseActive(_visit)
        src.close()
        return dst


class VectorFieldMerge:
    @staticmethod
    def Merge(oSource: VectorField, oTarget: VectorField) -> None:
        def _visit(p: tuple[float, float, float], v: tuple[float, float, float]) -> None:
            oTarget.SetValue(p, v)

        oSource.TraverseActive(_visit)



class AddVectorFieldToViewer:
    @staticmethod
    def AddToViewer(oViewer: IViewer, oField: VectorField, clr: ColorLike, nStep: int = 10, fArrow: float = 1.0, nGroup: int = 0) -> None:
        if nStep <= 0:
            raise ValueError("nStep must be > 0")
        color = to_rgba(clr)
        count = 0

        def _visit(pos: tuple[float, float, float], vec: tuple[float, float, float]) -> None:
            nonlocal count
            count += 1
            if count < nStep:
                return
            count = 0

            poly = PolyLine(color)
            poly.nAddVertex(pos)
            if abs(vec[0]) < 1e-12 and abs(vec[1]) < 1e-12 and abs(vec[2]) < 1e-12:
                poly.AddCross(fArrow)
            else:
                poly.nAddVertex((pos[0] + vec[0], pos[1] + vec[1], pos[2] + vec[2]))
                poly.AddArrow(fArrow)

            add_fn = getattr(oViewer, "Add", None)
            if callable(add_fn):
                add_fn(poly, nGroup)

        oField.TraverseActive(_visit)


class FieldUtils:
    SdfVisualizer = SdfVisualizer
    ActiveVoxelCounterScalar = ActiveVoxelCounterScalar
    SurfaceNormalFieldExtractor = SurfaceNormalFieldExtractor
    VectorFieldMerge = VectorFieldMerge
    AddVectorFieldToViewer = AddVectorFieldToViewer




from __future__ import annotations

from picogk import Voxels

from ..frames.local_frames import LocalFrame
from ..modulations import SurfaceModulation
from .base_cylinder import BaseCylinder
from .base_shape import BaseShape


class BaseCone(BaseShape):
	def __init__(self, oFrame: LocalFrame, fLength: float, fStartRadius: float, fEndRadius: float) -> None:
		super().__init__()
		self.m_fStartRadius = fStartRadius
		self.m_fEndRadius = fEndRadius
		self.m_oCyl = BaseCylinder.from_frame(oFrame, fLength=fLength)
		self.m_oCyl.SetRadius(SurfaceModulation(self.fGetLinearRadius))

	def fGetLinearRadius(self, fPhi: float, fLengthRatio: float) -> float:
		lr = max(0.0, min(1.0, fLengthRatio))
		return self.m_fStartRadius + lr * (self.m_fEndRadius - self.m_fStartRadius)

	def voxConstruct(self) -> Voxels:
		self.m_oCyl.SetTransformation(self.m_fnTrafo)
		return self.m_oCyl.voxConstruct()

	def oGetBaseCylinder(self) -> BaseCylinder:
		return self.m_oCyl


__all__ = ["BaseCone"]
